# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Filters and methods for control loops**

:Authors: **Birk Emil Karlsen-Bæck**, **Helga Timko**
'''

from __future__ import annotations
from typing import TYPE_CHECKING

import numpy
from scipy.special import comb
from blond.llrf.impulse_response import TravellingWaveCavity

# Set up logging
import logging

import matplotlib.pyplot as plt
import numpy as np
import numpy.linalg as npla
from scipy import signal as sgn
from scipy.constants import e

if TYPE_CHECKING:
    from blond.beam.profile import Profile

logger = logging.getLogger(__name__)


def polar_to_cartesian(amplitude, phase):
    """Convert data from polar to cartesian (I,Q) coordinates.

    Parameters
    ----------
    amplitude : float array
        Amplitude of signal
    phase : float array
        Phase of signal

    Returns
    -------
    complex array
        Signal with in-phase and quadrature (I,Q) components
    """

    logger.debug("Converting from polar to Cartesian")

    return amplitude * (np.cos(phase) + 1j * np.sin(phase))


def cartesian_to_polar(IQ_vector):
    """Convert data from Cartesian (I,Q) to polar coordinates.

    Parameters
    ----------
    IQ_vector : complex array
        Signal with in-phase and quadrature (I,Q) components

    Returns
    -------
    float array
        Amplitude of signal
    float array
        Phase of signal

    """

    logger.debug("Converting from Cartesian to polar")

    return np.absolute(IQ_vector), np.angle(IQ_vector)


def get_power_gen_i(I_gen_per_cav, Z_0):
    """RF generator power from generator current (physical, in [A]), for any
    f_r (and thus any tau)

    Parameters
    ----------
    I_gen_per_cav : complex array
        Generator current for a single cavity
    Z_0 : float

    Returns
    -------
    float array
        Absolute value of the generator power

    """
    return 0.5 * Z_0 * np.abs(I_gen_per_cav)**2


def modulator(signal, omega_i, omega_f, T_sampling, phi_0=0, dt=0):
    """Demodulate a signal from initial frequency to final frequency. The two
    frequencies should be close.

    Parameters
    ----------
    signal : float array
        Signal to be demodulated
    omega_i : float
        Initial revolution frequency [1/s] of signal (before demodulation)
    omega_f : float
        Final revolution frequency [1/s] of signal (after demodulation)
    T_sampling : float
        Sampling period (temporal bin size) [s] of the signal

    Returns
    -------
    float array
        Demodulated signal at f_final

    """

    if len(signal) < 2:
        # TypeError
        raise RuntimeError("ERROR in filters.py/demodulator: signal should" +
                           " be an array!")
    delta_phi = (omega_i - omega_f) * (T_sampling * np.arange(len(signal)) + dt)
    # Pre compute sine and cosine for speed up
    cs = np.cos(delta_phi + phi_0)
    sn = np.sin(delta_phi + phi_0)
    I_new = cs*signal.real - sn*signal.imag
    Q_new = sn*signal.real + cs*signal.imag

    return I_new + 1j * Q_new


def rf_beam_current(Profile: Profile, omega_c: float, T_rev: float, lpf: bool = True, downsample: dict = None,
                    external_reference: bool = True, dT: float = 0):
    r"""Function calculating the beam charge at the (RF) frequency, slice by
    slice. The charge distribution [C] of the beam is determined from the beam
    profile :math:`\lambda_i`, the particle charge :math:`q_p` and the real vs.
    macro-particle ratio :math:`N_{\mathsf{real}}/N_{\mathsf{macro}}`

    .. math::
        Q_i = \frac{N_{\mathsf{real}}}{N_{\mathsf{macro}}} q_p \lambda_i

    The total charge [C] in the beam is then

    .. math::
        Q_{\mathsf{tot}} = \sum_i{Q_i}

    The DC beam current [A] is the total number of charges per turn :math:`T_0`

    .. math:: I_{\mathsf{DC}} = \frac{Q_{\mathsf{tot}}}{T_0}

    The RF beam charge distribution [C] at a revolution frequency
    :math:`\omega_c` is the complex quantity

    .. math::
        \left( \begin{matrix} I_{rf,i} \\
        Q_{rf,i} \end{matrix} \right)
        = 2 Q_i \left( \begin{matrix} \cos(\omega_c t_i) \\
        \sin(\omega_c t_i)\end{matrix} \right) \, ,

    where :math:`t_i` are the time coordinates of the beam profile. After de-
    modulation, a low-pass filter at 20 MHz is applied.

    For multi-bunch cases, make sure that the real beam intensity is the total
    number of charges in the ring.

    Parameters
    ----------
    Profile : class
        A Profile type class
    omega_c : float
        Revolution frequency [1/s] at which the current should be calculated
    T_rev : float
        Revolution period [s] of the machine
    lpf : bool
        Apply low-pass filter; default is True
    downsample : dict
        Dictionary containing float value for 'Ts' sampling time and int value
        for 'points'. Will downsample the RF beam charge onto a coarse time
        grid with 'Ts' sampling time and 'points' points.
    external_reference : bool
        Option to include the changing external reference of the time-grid
    dT : float
        The shift in time due to shifting reference frames

    Returns
    -------
    complex array
        RF beam charge array [C] at 'frequency' omega_c, with the sampling time
        of the Profile object. To obtain current, divide by the sampling time
    (complex array)
        If time_coarse is specified, returns also the RF beam charge array [C]
        on the coarse time grid

    """

    # Convert from dimensionless to Coulomb/Ampères
    # Take into account macro-particle charge with real-to-macro-particle ratio
    charges = Profile.Beam.ratio * Profile.Beam.Particle.charge * e \
         * np.copy(Profile.n_macroparticles)
    logger.debug("Sum of particles: %d, total charge: %.4e C",
                 np.sum(Profile.n_macroparticles), np.sum(charges))
    logger.debug("DC current is %.4e A", np.sum(charges) / T_rev)

    # Mix with frequency of interest; remember factor 2 demodulation
    I_f = 2. * charges * np.cos(omega_c * Profile.bin_centers)
    Q_f = -2. * charges * np.sin(omega_c * Profile.bin_centers)

    # Pass through a low-pass filter
    if lpf is True:
        # Nyquist frequency 0.5*f_slices; cutoff at 20 MHz
        cutoff = 20.e6 * 2. * Profile.bin_size
        I_f = low_pass_filter(I_f, cutoff_frequency=cutoff)
        Q_f = low_pass_filter(Q_f, cutoff_frequency=cutoff)
    logger.debug("RF total current is %.4e A", np.fabs(np.sum(I_f)) / T_rev)

    charges_fine = I_f + 1j * Q_f
    if external_reference:
        # slippage in phase due to a non-integer harmonic number
        dphi = dT * omega_c
        # Total phase correction
        phase = dphi
        charges_fine = charges_fine * np.exp(1j * phase)

    if downsample:
        try:
            T_s = float(downsample['Ts'])
            n_points = int(downsample['points'])
        except Exception:
            raise RuntimeError('Downsampling input erroneous in rf_beam_current')

        # Find which index in fine grid matches index in coarse grid
        ind_fine = np.round((Profile.bin_centers + dT - np.pi / omega_c)/T_s)
        ind_fine = np.array(ind_fine, dtype=int)
        indices = np.where((ind_fine[1:] - ind_fine[:-1]) == 1)[0]

        # Pick total current within one coarse grid
        charges_coarse = np.zeros(n_points, dtype=complex)
        charges_coarse[ind_fine[0]] = np.sum(charges_fine[np.arange(indices[0])])
        for i in range(1, len(indices)):
            charges_coarse[i + ind_fine[0]] = np.sum(charges_fine[np.arange(indices[i - 1],
                                                                            indices[i])])

        return charges_fine, charges_coarse

    else:
        return charges_fine


def comb_filter(y, x, a):
    """Feedback comb filter.
    """

    return a * y + (1 - a) * x


def fir_filter_coefficients(n_taps, sampling_freq, cutoff_freq):
    """Band-stop type FIR filter from scipy
    http://docs.scipy.org

    Parameters
    ----------
    signal : complex array
        Signal to be filtered
    n_taps : int
        Number of taps, should be impair
    sampling_freq : float
        Sampling frequency [Hz]
    cutoff_freq : float
        Cutoff frequency [Hz]


    Returns
    -------
    ndarray
        FIR filter coefficients of length n_taps

    """

    fPass = cutoff_freq/sampling_freq

    return sgn.firwin(n_taps, [fPass], pass_zero=True)


def fir_filter_lhc_otfb_coeff(n_taps=63):
    '''FIR filter designed for the LHC OTFB, for a sampling frequency of
    40 MS/s, with 63 taps.

    Parameters
    ----------
    n_taps : int
        Number of taps. 63 for 40 MS/s or 15 for 10 MS/s

    Returns
    -------
    double array
        Coefficients of LHC-type FIR filter
    '''

    if n_taps == 15:
        coeff = [-0.0469, -0.016, 0.001, 0.0321, 0.0724, 0.1127, 0.1425,
                 0.1534, 0.1425, 0.1127, 0.0724, 0.0321, 0.001, -0.016, -0.0469]
    elif n_taps == 63:

        coeff = [-0.038636, -0.00687283, -0.00719296, -0.00733319, -0.00726159,
            -0.00694037, -0.00634775, -0.00548098, -0.00432789, -0.00288188,
            -0.0011339, 0.00090253, 0.00321323, 0.00577238, 0.00856464,
            0.0115605, 0.0147307, 0.0180265, 0.0214057, 0.0248156, 0.0282116,
            0.0315334, 0.0347311, 0.0377502, 0.0405575, 0.0431076, 0.0453585,
            0.047243, 0.0487253, 0.049782, 0.0504816, 0.0507121, 0.0504816,
            0.049782, 0.0487253, 0.047243, 0.0453585, 0.0431076, 0.0405575,
            0.0377502, 0.0347311, 0.0315334, 0.0282116, 0.0248156, 0.0214057,
            0.0180265, 0.0147307, 0.0115605, 0.00856464, 0.00577238, 0.00321323,
            0.00090253, -0.0011339, -0.00288188, -0.00432789, -0.00548098,
            -0.00634775, -0.00694037, -0.00726159, -0.00733319, -0.00719296,
            -0.00687283, -0.038636]
    else:
        raise ValueError("In LHC FIR filter, number of taps has to be 15 or 63")

    return coeff


def fir_filter(coeff, signal):
    '''Apply FIR filter on discrete time signal.

    Paramters
    ---------
    coeff : double array
        Coefficients of FIR filter with length of number of taps
    signal : complex or double array
        Input signal to be filtered

    Returns
    -------
    complex or double array
        Filtered signal of length len(signal) - len(coeff)
    '''

    n_taps = len(coeff)
    filtered_signal = np.zeros(len(signal) - n_taps)
    for i in range(n_taps, len(signal)):
        for k in range(n_taps):
            filtered_signal[i-n_taps] += coeff[k] * signal[i - k]

    return filtered_signal


def low_pass_filter(signal, cutoff_frequency=0.5):
    """Low-pass filter based on Butterworth 5th order digital filter from
    scipy,
    http://docs.scipy.org

    Parameters
    ----------
    signal : float array
        Signal to be filtered
    cutoff_frequency : float
        Cutoff frequency [1] corresponding to a 3 dB gain drop, relative to the
        Nyquist frequency of 1; default is 0.5

    Returns
    -------
    float array
        Low-pass filtered signal

    """

    b, a = sgn.butter(5, cutoff_frequency, 'low', analog=False)

    return sgn.filtfilt(b, a, signal)


def moving_average(x, N, x_prev=None):
    """Function to calculate the moving average (or running mean) of the input
    data.

    Parameters
    ----------
    x : float array
        Data to be smoothed
    N : int
        Window size in points
    x_prev : float array
        Data to pad with in front

    Returns
    -------
    float array
        Smoothed data array of size
            * len(x) - N + 1, if x_prev = None
            * len(x) + len(x_prev) - N + 1, if x_prev given

    """

    if x_prev is not None:
        # Pad in front with x_prev signal
        x = np.concatenate((x_prev, x))

    # based on https://stackoverflow.com/a/14314054
    mov_avg = np.cumsum(x)
    mov_avg[N:] = mov_avg[N:] - mov_avg[:-N]
    return mov_avg[N - 1:] / N


def moving_average_improved(x, N, x_prev=None):

    if x_prev is not None:
        x = np.concatenate((x_prev, x))

    mov_avg = sgn.fftconvolve(x, (1 / N) * np.ones(N), mode='full')[-x.shape[0]:]

    return mov_avg[:x.shape[0] - N + 1]


def H_cav(x, n_sections, x_prev=None):

    if x_prev is not None:
        x = np.concatenate((x_prev, x))

    if n_sections == 3:
        h = np.array([-0.04120219, -0.00765499, -0.00724786, -0.00600952, -0.00380694, -0.00067663,
                      0.00343537, 0.0084533, 0.01421418, 0.02071802, 0.02764441, 0.03476114,
                      0.04193753, 0.04882965, 0.05522681, 0.06083675, 0.0654471, 0.06887487,
                      0.07100091, 0.09043617, 0.07100091, 0.06887487, 0.0654471, 0.06083675,
                      0.05522681, 0.04882965, 0.04193753, 0.03476114, 0.02764441, 0.02071802,
                      0.01421418, 0.0084533, 0.00343537, -0.00067663, -0.00380694, -0.00600952,
                      -0.00724786, -0.00765499, -0.04120219])
    else:
        h = np.array([-0.0671217, 0.01355402, 0.01365686, 0.01444814, 0.01571424, 0.01766679,
                      0.01996413, 0.02251791, 0.02529718, 0.02817416, 0.03113348, 0.03398052,
                      0.03674144, 0.03924433, 0.04153931, 0.04344182, 0.04502165, 0.04612467,
                      0.04685122, 0.06409968, 0.04685122, 0.04612467, 0.04502165, 0.04344182,
                      0.04153931, 0.03924433, 0.03674144, 0.03398052, 0.03113348, 0.02817416,
                      0.02529718, 0.02251791, 0.01996413, 0.01766679, 0.01571424, 0.01444814,
                      0.01365686, 0.01355402, -0.0671217])

    resp = sgn.fftconvolve(x, h, mode='full')[-x.shape[0]:]

    return resp[:x.shape[0] - h.shape[0] + 1]


def smooth_step(x, x_min=0, x_max=1, N=1):
    """Function to make a smooth step.

    Parameters
    ----------
    x : float array
        Data to be smoothed
    x_min : float
        Minimum output value of step
    x_max : float
        Maximum output value of step
    N : int
        Order of smoothness

    Returns
    -------
    float array
        Smooth step of input signal
    Taken from: https://stackoverflow.com/questions/45165452/how-to-implement-a-smooth-clamp-function-in-python
    """
    x = np.clip((x - x_min) / (x_max - x_min), 0, 1)

    result = 0
    for n in range(0, N + 1):
        result += comb(N + n, n) * comb(2 * N + 1, N - n) * (-x) ** n

    result *= x ** (N + 1)

    return result


def feedforward_filter(TWC: TravellingWaveCavity, T_s, taps=None,
                       opt_output=False):
    """Function to design n-tap FIR filter for SPS TravellingWaveCavity.

    Parameters
    ----------
    TWC : TravellingWaveCavity
        TravellingWaveCavity type class
    T_s : float
        Sampling time [s]
    taps : int
        User-defined number of taps; default is None and number of taps is
        calculated from the filling time
    opt_output : bool
        When True, activates optional output; default is False

    Returns
    -------
    float array
        FIR filter coefficients
    int
        Optional output: Number of FIR filter taps
    int
        Optional output: Filling time in samples
    int
        Optional output: Fitting time in samples, n_filling, n_fit
    """

    # Filling time in samples
    n_filling = round(TWC.tau / T_s)
    logger.debug("Filling time in samples: %d", n_filling)

    # Number of FIR filter taps
    if taps is not None:
        n_taps = int(taps)
    else:
        n_taps = 2 * int(0.5 * n_filling) + 13  # 31

    if n_taps % 2 == 0:
        raise RuntimeError("Number of taps in feedforward filter must be odd!")
    logger.debug("Number of taps: %d", n_taps)

    # Fitting samples
    n_fit = int(n_taps + n_filling)
    logger.debug("Fitting samples: %d", n_fit)

    def V_real(t, tauf):
        output = np.zeros(t.shape)
        for i in range(len(t)):
            if t[i] < -tauf:
                output[i] = 0
            elif t[i] < 0:
                output[i] = (t[i]/tauf + 1)**2 / 2
            elif t[i] < tauf:
                output[i] = -1/2 * (t[i]/tauf)**2 + t[i]/tauf + 1/2
            else:
                output[i] = 1
        return output

    # Imaginary part
    def V_imag(t, tauf):
        output = np.zeros(t.shape)
        for i in range(len(t)):
            if t[i] < -tauf:
                output[i] = 0
            elif t[i] < 0:
                output[i] = -(t[i]/tauf + 1)**2 / 2
            elif t[i] < tauf:
                output[i] = -1/2 * (t[i]/tauf)**2 + t[i]/tauf - 1/2
            else:
                output[i] = 0
        return output

    Pfit_ = np.linspace(-(n_fit - 1) / 2, (n_fit - 1) / 2, n_fit)

    # Even symmetric part of beam loading
    Dvectoreven = V_real(Pfit_, n_filling)

    # Odd symmetric part of beam loading
    Dvectorodd = V_imag(Pfit_, n_filling)

    # Step response of FIR. (M1, N1) must be odd
    def Smatrix(M1, N1):
        output = np.zeros((M1, N1))
        for i in range(M1):
            for j in range(N1):
                if i - j >= (M1 - N1) / 2:
                    output[i, j] = 1
        return output

    # Response of symmetrix rectangle of length L. P must be odd
    def Rmatrix(P, L):
        output = np.zeros((P, P + L - 1))
        for i in range(output.shape[0]):
            for j in range(output.shape[1]):
                if i - j <= 0 and i - j > -L:
                    output[i, j] = 1
        return output

    def uniform_weighting(n, Nt):
        return 1

    def Weigthing(Nt, weighting_function):
        output = np.zeros((Nt, Nt))
        for i in range(output.shape[0]):
            for j in range(output.shape[1]):
                if i == j:
                    output[i, j] = weighting_function(i, Nt)
        return output

    def EvenMatrix(Nt):
        output = np.zeros((Nt, (Nt + 1) // 2))
        for i in range(output.shape[0]):
            for j in range(output.shape[1]):
                if i + j == (Nt + 0) // 2:
                    output[i, j] = 1
                elif i - j == (Nt - 1) // 2:
                    output[i, j] = 1
        return output

    def OddMatrix(Nt):
        output = np.zeros((Nt, (Nt - 1) // 2))
        for i in range(output.shape[0]):
            for j in range(output.shape[1]):
                if i + j == (Nt - 2) // 2:
                    output[i, j] = -1
                elif i - j == (Nt + 1) // 2:
                    output[i, j] = 1
        return output

    def Hoptreal(Nt, L, P, Dvectoreven):
        output = EvenMatrix(Nt).T @ Smatrix(P + L - 1, Nt).T @ Rmatrix(P, L).T @ Dvectoreven

        matrix1 = EvenMatrix(Nt)
        matrix1 = Rmatrix(P, L).T @ Weigthing(P, uniform_weighting) @ Rmatrix(P, L) @ Smatrix(P + L - 1, Nt) @ matrix1
        matrix1 = EvenMatrix(Nt).T @ Smatrix(P + L - 1, Nt).T @ matrix1
        matrix1 = npla.inv(matrix1)

        return matrix1 @ output

    def Hoptimag(Nt, L, P, Dvectorodd):
        output = OddMatrix(Nt).T @ Smatrix(P + L - 1, Nt).T @ Rmatrix(P, L).T @ Weigthing(P, uniform_weighting) \
                 @ Dvectorodd
        matrix1 = OddMatrix(Nt)
        matrix1 = Rmatrix(P, L).T @ Weigthing(P, uniform_weighting) @ Rmatrix(P, L) @ Smatrix(P + L - 1, Nt) @ matrix1
        matrix1 = OddMatrix(Nt).T @ Smatrix(P + L - 1, Nt).T @ matrix1
        matrix1 = npla.inv(matrix1)

        return matrix1 @ output

    def Hopteven(Nt, L, P, Dvectoreven):
        return np.concatenate([Hoptreal(Nt, L, P, Dvectoreven)[1:][::-1], Hoptreal(Nt, L, P, Dvectoreven)])

    def Hoptodd(Nt, L, P, Dvectorodd):
        output = np.concatenate([-Hoptimag(Nt, L, P, Dvectorodd)[::-1], np.array([0])])
        return np.concatenate([output, Hoptimag(Nt, L, P, Dvectorodd)])

    h_ff = Hopteven(n_taps, n_filling, n_fit, Dvectoreven) + Hoptodd(n_taps, n_filling, n_fit, Dvectorodd)

    if opt_output:
        return h_ff, n_taps, n_filling, n_fit
    else:
        return h_ff


feedforward_filter_TWC3 = np.array(
    [-0.00760838, 0.01686764, 0.00205761, 0.00205761,
     0.00205761, 0.00205761, -0.03497942, 0.00205761,
     0.00205761, 0.00205761, 0.00205761, -0.0053474,
     0.00689061, 0.00308642, 0.00308642, 0.00308642,
     0.00308642, 0.00308642, -0.00071777, 0.01152024,
     0.00411523, 0.00411523, 0.00411523, 0.00411523,
     0.03806584, -0.00205761, -0.00205761, -0.00205761,
     -0.00205761, -0.01686764, 0.00760838])

feedforward_filter_TWC4 = np.array(
    [0.01050256, -0.0014359, 0.00106667, 0.00106667,
     0.00106667, -0.01226667, -0.01226667, 0.00106667,
     0.00106667, 0.00106667, 0.00231795, -0.00365128,
     0.0016, 0.0016, 0.0016, 0.0016,
     0.0016, 0.0016, 0.0016, 0.0016,
     0.0016, 0.0016, 0.0016, 0.0016,
     0.0016, 0.00685128, 0.00088205, 0.00213333,
     0.00213333, 0.00213333, 0.01506667, 0.01266667,
     -0.00106667, -0.00106667, -0.00106667, 0.0014359,
     -0.01050256])

feedforward_filter_TWC5 = np.array(
    [0.01802423, -0.01004643,  0.00069372,  0.00069372,  0.00069372, -0.01005897,
     -0.01005897,  0.00069372,  0.00069372,  0.00069372,  0.0060638,  -0.00797153,
     0.00104058,  0.00104058,  0.00104058,  0.00104058,  0.00104058,  0.00104058,
     0.00104058,  0.00104058,  0.00104058,  0.00104058,  0.00104058,  0.00104058,
     0.00104058,  0.00104058,  0.00104058,  0.00104058,  0.00104058,  0.00104058,
     0.00104058,  0.0100527,  -0.00398263,  0.00138744,  0.00138744,  0.00138744,
     0.01187999,  0.01031911, -0.00069372, -0.00069372, -0.00069372,  0.01004643,
     -0.01802423])


def plot_frequency_response(b, a=1):
    """Plotting the frequency response of a filter with coefficients a, b."""

    w, H = sgn.freqz(b, a)

    plt.subplot(211)
    plt.plot(2 * w / np.max(w), np.absolute(H))
    plt.ylabel('Amplitude [linear]')
    plt.xlabel(r'Frequency w.r.t. sampling frequency')
    plt.title(r'Frequency response')

    plt.subplot(212)
    phase = np.unwrap(np.angle(H))
    plt.plot(w / max(w), phase)
    plt.ylabel('Phase [radians]')
    plt.xlabel(r'Frequency w.r.t. sampling frequency')
    plt.title(r'Phase response')
    plt.subplots_adjust(hspace=0.5)

    plt.show()



# coding: utf-8
# Copyright 2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

"""Module containing the fundamental beam class with methods to compute beam
statistics

:Authors: **Danilo Quartullo**, **Helga Timko**, **ALexandre Lasheen**

"""

from __future__ import division

import itertools as itl

import numpy as np
from scipy.constants import c, e, epsilon_0, hbar, m_e, m_p, physical_constants

from ..trackers.utilities import is_in_separatrix
from ..utils import bmath as bm
from ..utils import exceptions as blExcept

m_mu = physical_constants['muon mass'][0]

class Particle:
    r"""Class containing basic parameters, e.g. mass, of the particles to be tracked.

    The following particles are already implemented: proton, electron, positron

    Parameters
    ----------
    user_mass : float
        Energy equivalent of particle rest mass in eV
    user_charge : float
        Particle charge in units of the elementary charge
    user_decay_rate : float
        Particle decay rate in units of 1/s 

    Attributes
    ----------
    mass : float
        Energy equivalent of particle rest mass in eV.
    charge : float
        Particle charge in units of the elementary charge.
    decay_rate : float
        Inverse of the particle decay time (decay time in s)
    radius_cl : float
        Classical particle radius in :math:`m`.
    c_gamma : float
        Sand's radiation constant :math:`C_\gamma` in :math:`m / eV^3`.
    c_q : float
        Quantum radiation constant :math:`C_q` in :math:`m`.

    Examples
    --------
    >>> from blond.beam.beam import Proton
    >>> particle = Proton()

    Usually, a `Particle` is used to construct a :class:`~blond.input_parameters.ring.Ring` object,
    e.g.

    >>> Ring(circumference, momentum_compaction, sync_momentum, Proton())

    """

    def __init__(self, user_mass, user_charge, user_decay_rate = 0):

        if user_mass > 0.:
            self.mass = float(user_mass)
            self.charge = float(user_charge)
        else:
            # MassError
            raise RuntimeError('ERROR: Particle mass not recognized!')
            
        if user_decay_rate >= 0.:
            self.decay_rate = float(user_decay_rate)
            
        else:
            # MassError
            raise RuntimeError('ERROR: Invalide particle decay rate!')
            

        # classical particle radius [m]
        self.radius_cl = 0.25 / (np.pi * epsilon_0) * \
            e**2 * self.charge**2 / (self.mass * e)

        # Sand's radiation constant [ m / eV^3]
        self.c_gamma = 4 * np.pi / 3 * self.radius_cl / self.mass**3

        # Quantum radiation constant [m]
        self.c_q = (55.0 / (32.0 * np.sqrt(3.0)) * hbar * c / (self.mass * e))


class Proton(Particle):
    """ Implements a proton `Particle`.
    """

    def __init__(self):

        Particle.__init__(self, m_p * c**2 / e, 1)


class Electron(Particle):
    """ Implements an electron `Particle`.
    """

    def __init__(self):
        Particle.__init__(self, m_e * c**2 / e, -1)


class Positron(Particle):
    """ Implements a positron `Particle`.
    """

    def __init__(self):

        Particle.__init__(self, m_e * c**2 / e, 1)


class MuPlus(Particle):
    """ Implements a muon+ `Particle`.
    """ 
    def __init__(self):
        Particle.__init__(self, m_mu * c**2 / e, 1, float(1/2.1969811e-6))


class MuMinus(Particle):
    """ Implements a muon- `Particle`.
    """ 
    def __init__(self):        
        Particle.__init__(self, m_mu * c**2 / e, -1, float(1/2.1969811e-6))
        

class Beam:
    r"""Class containing the beam properties.

    This class containes the beam coordinates (dt, dE) and the beam properties.

    The beam coordinate 'dt' is defined as the particle arrival time to the RF
    station w.r.t. the reference time that is the sum of turns. The beam
    coordiate 'dE' is defined as the particle energy offset w.r.t. the
    energy of the synchronous particle.

    The class creates a beam with zero dt and dE, see distributions to match
    a beam with respect to the RF and intensity effects.

    Parameters
    ----------
    Ring : Ring
        Used to import different quantities such as the mass and the energy.
    n_macroparticles : int
        total number of macroparticles.
    intensity : float
        total intensity of the beam (in number of charge).

    Attributes
    ----------
    mass : float
        mass of the particle [eV].
    charge : int
        integer charge of the particle [e].
    beta : float
        relativistic velocity factor [].
    gamma : float
        relativistic mass factor [].
    energy : float
        energy of the synchronous particle [eV].
    momentum : float
        momentum of the synchronous particle [eV].
    dt : numpy_array, float
        beam arrival times with respect to synchronous time [s].
    dE : numpy_array, float
        beam energy offset with respect to the synchronous particle [eV].
    mean_dt : float
        average beam arrival time [s].
    mean_dE : float
        average beam energy offset [eV].
    sigma_dt : float
        standard deviation of beam arrival time [s].
    sigma_dE : float
        standard deviation of beam energy offset [eV].
    intensity : float
        total intensity of the beam in number of charges [].
    n_macroparticles : int
        total number of macroparticles in the beam [].
    ratio : float
        ratio intensity per macroparticle [].
    id : numpy_array, int
        unique macro-particle ID number; zero if particle is 'lost'.

    See Also
    ---------
    distributions.matched_from_line_density:
        match a beam with a given bunch profile.
    distributions.matched_from_distribution_function:
        match a beam with a given distribution function in phase space.

    Examples
    --------
    >>> from input_parameters.ring import Ring
    >>> from beam.beam import Beam
    >>>
    >>> n_turns = 10
    >>> C = 100
    >>> eta = 0.03
    >>> momentum = 26e9
    >>> ring = Ring(n_turns, C, eta, momentum, 'proton')
    >>> n_macroparticle = 1e6
    >>> intensity = 1e11
    >>>
    >>> my_beam = Beam(ring, n_macroparticle, intensity)
    """

    def __init__(self, Ring, n_macroparticles, intensity):

        self.Particle = Ring.Particle
        self.beta = Ring.beta[0][0]
        self.gamma = Ring.gamma[0][0]
        self.energy = Ring.energy[0][0]
        self.momentum = Ring.momentum[0][0]
        self.dt = np.zeros([int(n_macroparticles)], dtype=bm.precision.real_t)
        self.dE = np.zeros([int(n_macroparticles)], dtype=bm.precision.real_t)
        self.mean_dt = 0.
        self.mean_dE = 0.
        self.sigma_dt = 0.
        self.sigma_dE = 0.
        self.intensity = float(intensity)
        self.n_macroparticles = int(n_macroparticles)
        self.ratio = self.intensity / self.n_macroparticles
        self.id = np.arange(1, self.n_macroparticles + 1, dtype=int)
        self.epsn_rms_l = 0.
        # For MPI
        self.n_total_macroparticles_lost = 0
        self.n_total_macroparticles = n_macroparticles
        self.is_splitted = False
        self._sumsq_dt = 0.
        self._sumsq_dE = 0.
        # For GPU
        self._device = 'CPU'

    @property
    def n_macroparticles_lost(self):
        '''Number of lost macro-particles, defined as @property.

        Returns
        -------
        n_macroparticles_lost : int
            number of macroparticles lost.

        '''
        return self.n_macroparticles - self.n_macroparticles_alive

    @property
    def n_macroparticles_alive(self):
        '''Number of transmitted macro-particles, defined as @property.

        Returns
        -------
        n_macroparticles_alive : int
            number of macroparticles not lost.

        '''

        return bm.count_nonzero(self.id)

    def eliminate_lost_particles(self):
        """Eliminate lost particles from the beam coordinate arrays
        """

        indexalive = np.where(self.id != 0)[0]
        if len(indexalive) > 0:
            self.dt = np.ascontiguousarray(
                self.dt[indexalive], dtype=bm.precision.real_t)
            self.dE = np.ascontiguousarray(
                self.dE[indexalive], dtype=bm.precision.real_t)
            self.n_macroparticles = len(self.dt)
            self.id = np.arange(1, self.n_macroparticles + 1, dtype=int)
        else:
            # AllParticlesLost
            raise RuntimeError("ERROR in Beams: all particles lost and" +
                               " eliminated!")

    def statistics(self):
        r'''
        Calculation of the mean and standard deviation of beam coordinates,
        as well as beam emittance using different definitions.
        Take no arguments, statistics stored in

        - mean_dt
        - mean_dE
        - sigma_dt
        - sigma_dE
        '''

        # Statistics only for particles that are not flagged as lost
        itemindex = bm.nonzero(self.id)[0]
        self.mean_dt = bm.mean(self.dt[itemindex])
        self.sigma_dt = bm.std(self.dt[itemindex])
        self._sumsq_dt = bm.dot(self.dt[itemindex], self.dt[itemindex])
        # self.min_dt = bm.min(self.dt[itemindex])
        # self.max_dt = bm.max(self.dt[itemindex])

        self.mean_dE = bm.mean(self.dE[itemindex])
        self.sigma_dE = bm.std(self.dE[itemindex])
        self._sumsq_dE = bm.dot(self.dE[itemindex], self.dE[itemindex])

        # self.min_dE = bm.min(self.dE[itemindex])
        # self.max_dE = bm.max(self.dE[itemindex])

        # R.m.s. emittance in Gaussian approximation
        self.epsn_rms_l = np.pi * self.sigma_dE * self.sigma_dt  # in eVs

    def losses_separatrix(self, Ring, RFStation):
        '''Beam losses based on separatrix.

        Set to 0 all the particle's id not in the separatrix anymore.

        Parameters
        ----------
        Ring : Ring
            Used to call the function is_in_separatrix.
        RFStation : RFStation
            Used to call the function is_in_separatrix.
        '''

        lost_index = is_in_separatrix(Ring, RFStation, self,
                                      self.dt, self.dE) == False

        self.id[lost_index] = 0

    def losses_longitudinal_cut(self, dt_min, dt_max):
        '''Beam losses based on longitudinal cuts.

        Set to 0 all the particle's id with dt not in the interval
        (dt_min, dt_max).

        Parameters
        ----------
        dt_min : float
            minimum dt.
        dt_max : float
            maximum dt.
        '''

        lost_index = (self.dt < dt_min) | (self.dt > dt_max)
        self.id[lost_index] = 0

    def losses_energy_cut(self, dE_min, dE_max):
        '''Beam losses based on energy cuts, e.g. on collimators.

        Set to 0 all the particle's id with dE not in the interval (dE_min, dE_max).

        Parameters
        ----------
        dE_min : float
            minimum dE.
        dE_max : float
            maximum dE.
        '''

        lost_index = (self.dE < dE_min) | (self.dE > dE_max)
        self.id[lost_index] = 0

    def losses_below_energy(self, dE_min):
        '''Beam losses based on lower energy cut.

        Set to 0 all the particle's id with dE below dE_min.

        Parameters
        ----------
        dE_min : float
            minimum dE.
        '''

        lost_index = (self.dE < dE_min)
        self.id[lost_index] = 0

    def particle_decay(self, time: float) -> None:
        '''Decreases beam inensity due to the particle decay

        Sets the ratio to a lower value if the particle can decay. Number of macroparticles remains unchanged.

        Parameters
        ----------
        time : float
            time in seconds, which is used to determine the fraction of the
            particle decay
        '''
        self.ratio *= np.exp(-time * self.Particle.decay_rate / (self.gamma))
       

    def add_particles(self, new_particles):
        '''
        Method to add array of new particles to beam object
        New particles are given id numbers sequential from last id of this beam

        Parameters
        ----------
        new_particles : array-like
            (2, n) array of (dt, dE) for new particles
        '''

        try:
            newdt = new_particles[0]
            newdE = new_particles[1]
            if len(newdt) != len(newdE):
                raise blExcept.ParticleAdditionError(
                    "new_particles must have equal number of time and energy coordinates")
        except TypeError:
            raise blExcept.ParticleAdditionError(
                "new_particles shape must be (2, n)")

        nNew = len(newdt)

        self.id = bm.concatenate((self.id, bm.arange(self.n_macroparticles + 1,
                                                     self.n_macroparticles
                                                     + nNew + 1, dtype=int)))
        self.n_macroparticles += nNew

        self.dt = bm.concatenate((self.dt, newdt))
        self.dE = bm.concatenate((self.dE, newdE))

    def add_beam(self, other_beam):
        '''
        Method to add the particles from another beam to this beam
        New particles are given id numbers sequential from last id of this beam
        Particles with id == 0 keep id == 0 and are included in addition

        Parameters
        ----------
        other_beam : blond beam object
        '''

        if not isinstance(other_beam, type(self)):
            raise TypeError("add_beam method requires a beam object as input")

        self.dt = bm.concatenate((self.dt, other_beam.dt))
        self.dE = bm.concatenate((self.dE, other_beam.dE))

        counter = itl.count(self.n_macroparticles + 1)
        newids = bm.zeros(other_beam.n_macroparticles)

        for i in range(other_beam.n_macroparticles):
            if other_beam.id[i]:
                newids[i] = next(counter)
            else:
                next(counter)

        self.id = bm.concatenate((self.id, newids))
        self.n_macroparticles += other_beam.n_macroparticles

    def __iadd__(self, other):
        '''
        Initialisation of in place addition calls add_beam(other) if other
        is a blond beam object, calls add_particles(other) otherwise

        Parameters
        ----------
        other : blond beam object or (2, n) array
        '''

        if isinstance(other, type(self)):
            self.add_beam(other)
            return self

        self.add_particles(other)
        return self

    def split(self, random=False, fast=False):
        '''
        MPI ONLY ROUTINE: Splits the beam equally among the workers for
        MPI processing.

        Parameters
        ----------
        random : boolean
            Shuffle the beam before splitting, to be used with the
            approximation methonds.
        fast : boolean
            If true, it assumes that every worker has already a copy of the
            beam so only the particle ids are distributed.
            If false, all the coordinates are distributed by the master to all
            the workers.
        '''

        if not bm.in_mpi():
            raise RuntimeError(
                'ERROR: Cannot use this routine unless in MPI Mode')

        from ..utils.mpi_config import WORKER
        if WORKER.is_master and random:
            bm.random.shuffle(self.id)
            if not fast:
                self.dt = self.dt[self.id - 1]
                self.dE = self.dE[self.id - 1]

        self.id = WORKER.scatter(self.id)
        if fast:
            self.dt = bm.ascontiguousarray(self.dt[self.id - 1])
            self.dE = bm.ascontiguousarray(self.dE[self.id - 1])
        else:
            self.dt = WORKER.scatter(self.dt)
            self.dE = WORKER.scatter(self.dE)

        assert (len(self.dt) == len(self.dE) and len(self.dt) == len(self.id))

        self.n_macroparticles = len(self.dt)
        self.is_splitted = True

    def gather(self, all_gather=False):
        '''
        MPI ONLY ROUTINE: Gather the beam coordinates to the master or all workers.

        Parameters
        ----------
        all_gather : boolean
            If true, every worker will get a copy of the whole beam coordinates.
            If false, only the master will gather the coordinates.
        '''
        if not bm.in_mpi():
            raise RuntimeError(
                'ERROR: Cannot use this routine unless in MPI Mode')
        from ..utils.mpi_config import WORKER

        if all_gather:
            self.dt = WORKER.allgather(self.dt)
            self.dE = WORKER.allgather(self.dE)
            self.id = WORKER.allgather(self.id)
            self.is_splitted = False
        else:
            self.dt = WORKER.gather(self.dt)
            self.dE = WORKER.gather(self.dE)
            self.id = WORKER.gather(self.id)
            if WORKER.is_master:
                self.is_splitted = False

        self.n_macroparticles = len(self.dt)

    def gather_statistics(self, all_gather=False):
        '''
        MPI ONLY ROUTINE: Gather beam statistics.

        Parameters
        ----------
        all_gather : boolean
            if true, all workers will gather the beam stats.
            If false, only the master will get the beam stats.
        '''
        if not bm.in_mpi():
            raise RuntimeError(
                'ERROR: Cannot use this routine unless in MPI Mode')

        from ..utils.mpi_config import WORKER
        if all_gather:

            self.mean_dt = WORKER.allreduce(
                np.array([self.mean_dt]), operator='mean')[0]

            self.mean_dE = WORKER.allreduce(
                np.array([self.mean_dE]), operator='mean')[0]

            self.n_total_macroparticles_lost = WORKER.allreduce(
                np.array([self.n_macroparticles_lost]), operator='sum')[0]

            # self.n_total_macroparticles_alive = WORKER.allreduce(
            # np.array([self.n_macroparticles_alive]), operator='sum')[0]

            self.sigma_dt = WORKER.allreduce(
                np.array([self._sumsq_dt]), operator='sum')[0]
            self.sigma_dt = np.sqrt(
                self.sigma_dt / (self.n_total_macroparticles -
                                 self.n_total_macroparticles_lost)
                - self.mean_dt**2)

            self.sigma_dE = WORKER.allreduce(
                np.array([self._sumsq_dE]), operator='sum')[0]
            self.sigma_dE = np.sqrt(
                self.sigma_dE / (self.n_total_macroparticles -
                                 self.n_total_macroparticles_lost)
                - self.mean_dE**2)

        else:
            self.mean_dt = WORKER.reduce(
                np.array([self.mean_dt]), operator='mean')[0]

            self.mean_dE = WORKER.reduce(
                np.array([self.mean_dE]), operator='mean')[0]

            self.n_total_macroparticles_lost = WORKER.reduce(
                np.array([self.n_macroparticles_lost]), operator='sum')[0]

            self.sigma_dt = WORKER.reduce(
                np.array([self._sumsq_dt]), operator='sum')[0]
            self.sigma_dt = np.sqrt(
                self.sigma_dt / (self.n_total_macroparticles -
                                 self.n_total_macroparticles_lost)
                - self.mean_dt**2)

            self.sigma_dE = WORKER.reduce(
                np.array([self._sumsq_dE]), operator='sum')[0]
            self.sigma_dE = np.sqrt(
                self.sigma_dE / (self.n_total_macroparticles -
                                 self.n_total_macroparticles_lost)
                - self.mean_dE**2)

    def gather_losses(self, all_gather=False):
        '''
        MPI ONLY ROUTINE: Gather beam losses.

        Parameters
        ----------
        all_gather : boolean
            if true, all workers will gather the beam stats.
            If false, only the master will get the beam stats.
        '''
        if not bm.in_mpi():
            raise RuntimeError(
                'ERROR: Cannot use this routine unless in MPI Mode')

        from ..utils.mpi_config import WORKER

        if all_gather:
            temp = WORKER.allgather(np.array([self.n_macroparticles_lost]))
            self.n_total_macroparticles_lost = np.sum(temp)
        else:
            temp = WORKER.gather(np.array([self.n_macroparticles_lost]))
            self.n_total_macroparticles_lost = np.sum(temp)

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        import cupy as cp
        self.dE = cp.array(self.dE)
        self.dt = cp.array(self.dt)
        self.id = cp.array(self.id)

        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        import cupy as cp
        self.dE = cp.asnumpy(self.dE)
        self.dt = cp.asnumpy(self.dt)
        self.id = cp.asnumpy(self.id)

        # to make sure it will not be called again
        self._device = 'CPU'




# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to compute intensity effects**

:Authors: **Juan F. Esteban Mueller**, **Danilo Quartullo**,
          **Alexandre Lasheen**, **Markus Schwarz**
'''

from __future__ import division, print_function

from builtins import range

import numpy as np
from scipy.constants import e

from ..toolbox.next_regular import next_regular
from ..utils import bmath as bm


class TotalInducedVoltage:
    r"""
    Object gathering all the induced voltage contributions. The input is a
    list of objects able to compute induced voltages (InducedVoltageTime,
    InducedVoltageFreq, InductiveImpedance). All the induced voltages will
    be summed in order to reduce the computing time. All the induced
    voltages should have the same slicing resolution.

    Parameters
    ----------
    Beam : object
        Beam object
    Profile : object
        Profile object
    induced_voltage_list : object list
        List of objects for which induced voltages have to be calculated

    Attributes
    ----------
    beam : object
        Copy of the Beam object in order to access the beam info
    profile : object
        Copy of the Profile object in order to access the profile info
    induced_voltage_list : object list
        List of objects for which induced voltages have to be calculated
    induced_voltage : float array
        Array to store the computed induced voltage [V]
    time_array : float array
        Time array corresponding to induced_voltage [s]
    """

    def __init__(self, Beam, Profile, induced_voltage_list):
        """
        Constructor.
        """
        # Copy of the Beam object in order to access the beam info.
        self.beam = Beam

        # Copy of the Profile object in order to access the profile info.
        self.profile = Profile

        # Induced voltage list.
        self.induced_voltage_list = induced_voltage_list

        # Induced voltage from the sum of the wake sources in V
        self.induced_voltage = np.zeros(
            int(self.profile.n_slices), dtype=bm.precision.real_t, order='C')

        # Time array of the wake in s
        self.time_array = self.profile.bin_centers

    def reprocess(self):
        """
        Reprocess the impedance contributions. To be run when profile changes
        """

        for induced_voltage_object in self.induced_voltage_list:
            induced_voltage_object.process()

    def induced_voltage_sum(self):
        """
        Method to sum all the induced voltages in one single array.
        """
        # For MPI, to avoid calulating beam spectrum multiple times
        beam_spectrum_dict = {}
        temp_induced_voltage = 0

        for induced_voltage_object in self.induced_voltage_list:
            induced_voltage_object.induced_voltage_generation(
                beam_spectrum_dict)
            temp_induced_voltage += \
                induced_voltage_object.induced_voltage[:self.profile.n_slices]

        self.induced_voltage = temp_induced_voltage.astype(
            dtype=bm.precision.real_t, order='C', copy=False)

    def track(self):
        """
        Track method to apply the induced voltage kick on the beam.
        """

        self.induced_voltage_sum()
        bm.linear_interp_kick(dt=self.beam.dt, dE=self.beam.dE,
                              voltage=self.induced_voltage,
                              bin_centers=self.profile.bin_centers,
                              charge=self.beam.Particle.charge,
                              acceleration_kick=0.)

    def track_ghosts_particles(self, ghostBeam):

        bm.linear_interp_kick(dt=ghostBeam.dt, dE=ghostBeam.dE,
                              voltage=self.induced_voltage,
                              bin_centers=self.profile.bin_centers,
                              charge=self.beam.Particle.charge,
                              acceleration_kick=0.)

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        if recursive:
            # transfer recursively objects
            for obj in self.induced_voltage_list:
                obj.to_gpu()

        import cupy as cp
        self.induced_voltage = cp.array(self.induced_voltage)
        self.time_array = cp.array(self.time_array)

        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        if recursive:
            # transfer recursively objects
            for obj in self.induced_voltage_list:
                obj.to_cpu()

        import cupy as cp
        self.induced_voltage = cp.asnumpy(self.induced_voltage)
        self.time_array = cp.asnumpy(self.time_array)
        # to make sure it will not be called again
        self._device = 'CPU'


class _InducedVoltage:
    r"""
    Induced voltage parent class. Only for internal use (inheritance), not to
    be directly instanciated.

    Parameters
    ----------
    Beam : object
        Beam object
    Profile : object
        Profile object
    frequency_resolution : float, optional
        Frequency resolution of the impedance [Hz]
    wake_length : float, optional
        Wake length [s]
    multi_turn_wake : boolean, optional
        Multi-turn wake enable flag
    mtw_mode : boolean, optional
        Multi-turn wake mode can be 'freq' or 'time' (default)
    RFParams : object, optional
        RFStation object for turn counter and revolution period
    use_regular_fft : boolean
        use the next_regular function to ensure regular number for FFT
        calculations (default is True for efficient calculations, for
        better control of the sampling frequency False is preferred)

    Attributes
    ----------
    beam : object
        Copy of the Beam object in order to access the beam info
    profile : object
        Copy of the Profile object in order to access the profile info
    induced_voltage : float array
        Induced voltage from the sum of the wake sources in V
    wake_length_input : float
        Wake length [s]
    frequency_resolution_input : float
        Frequency resolution of the impedance [Hz]
    RFParams : object
        RFStation object for turn counter and revolution period
    multi_turn_wake : boolean
        Multi-turn wake enable flag
    mtw_mode : boolean
        Multi-turn wake mode can be 'freq' or 'time' (default)
    use_regular_fft : boolean
        User set value to use (default) or not regular numbers for FFTs
    """

    def __init__(self, Beam, Profile, frequency_resolution=None,
                 wake_length=None, multi_turn_wake=False, mtw_mode='time',
                 RFParams=None, use_regular_fft=True):

        # Beam object in order to access the beam info
        self.beam = Beam

        # Profile object in order to access the profile info
        self.profile = Profile

        # Induced voltage from the sum of the wake sources in V
        self.induced_voltage = 0

        # Wake length in s (optional)
        self.wake_length_input = wake_length

        # Frequency resolution of the impedance (optional)
        self.frequency_resolution_input = frequency_resolution

        # Use regular numbers for fft (optional)
        self.use_regular_fft = use_regular_fft

        # RFStation object for turn counter and revolution period
        self.RFParams = RFParams

        # Multi-turn wake enable flag
        self.multi_turn_wake = multi_turn_wake

        # Multi-turn wake mode can be 'freq' or 'time' (default). If 'freq'
        # is used, each turn the induced voltage of previous turns is shifted
        # in the frequency domain. For 'time', a linear interpolation is used.
        self.mtw_mode = mtw_mode

        self.process()

    def process(self):
        """
        Reprocess the impedance contributions. To be run when profile changes
        """

        if (self.wake_length_input is not None
                and self.frequency_resolution_input is None):
            # Number of points of the induced voltage array
            self.n_induced_voltage = int(
                np.ceil(self.wake_length_input / self.profile.bin_size))
            if self.n_induced_voltage < self.profile.n_slices:
                # WakeLengthError
                raise RuntimeError('Error: too short wake length. ' +
                                   'Increase it above {0:1.2e} s.'.format(
                                       self.profile.n_slices * self.profile.bin_size))
            # Wake length in s, rounded up to the next multiple of bin size
            self.wake_length = self.n_induced_voltage * self.profile.bin_size
        elif (self.frequency_resolution_input is not None
                and self.wake_length_input is None):
            self.n_induced_voltage = int(
                np.ceil(1 / (self.profile.bin_size * self.frequency_resolution_input)))
            if self.n_induced_voltage < self.profile.n_slices:
                # FrequencyResolutionError
                raise RuntimeError('Error: too large frequency_resolution. ' +
                                   'Reduce it below {0:1.2e} Hz.'.format(
                                       1 / (self.profile.cut_right - self.profile.cut_left)))
            self.wake_length = self.n_induced_voltage * self.profile.bin_size
            # Frequency resolution in Hz
        elif (self.wake_length_input is None
                and self.frequency_resolution_input is None):
            # By default the wake_length is the slicing frame length
            self.wake_length = (self.profile.cut_right -
                                self.profile.cut_left)
            self.n_induced_voltage = self.profile.n_slices
        else:
            raise RuntimeError('Error: only one of wake_length or ' +
                               'frequency_resolution can be specified.')

        if self.multi_turn_wake:
            # Number of points of the memory array for multi-turn wake
            self.n_mtw_memory = self.n_induced_voltage

            self.front_wake_buffer = 0

            if self.mtw_mode == 'freq':
                # In frequency domain, an extra buffer for a revolution turn is
                # needed due to the circular time shift in frequency domain
                self.buffer_size = \
                    np.ceil(np.max(self.RFParams.t_rev) /
                            self.profile.bin_size)
                # Extending the buffer to reduce the effect of the front wake
                self.buffer_size += \
                    np.ceil(np.max(self.buffer_extra) / self.profile.bin_size)
                self.n_mtw_memory += int(self.buffer_size)
                # Using next regular for FFTs speedup
                if self.use_regular_fft:
                    self.n_mtw_fft = next_regular(self.n_mtw_memory)
                else:
                    self.n_mtw_fft = self.n_mtw_memory
                # Frequency and omega arrays
                self.freq_mtw = \
                    bm.rfftfreq(self.n_mtw_fft, d=self.profile.bin_size)
                self.omegaj_mtw = 2.0j * np.pi * self.freq_mtw
                # Selecting time-shift method
                self.shift_trev = self.shift_trev_freq
            else:
                # Selecting time-shift method
                self.shift_trev = self.shift_trev_time
                # Time array
                self.time_mtw = np.linspace(0, self.wake_length,
                                            self.n_mtw_memory, endpoint=False,
                                            dtype=bm.precision.real_t)

            # Array to add and shift in time the multi-turn wake over the turns
            self.mtw_memory = np.zeros(self.n_mtw_memory,
                                       dtype=bm.precision.real_t, order='C')

            # Select induced voltage generation method to be used
            self.induced_voltage_generation = self.induced_voltage_mtw
        else:
            self.induced_voltage_generation = self.induced_voltage_1turn

    def induced_voltage_1turn(self, beam_spectrum_dict={}):
        """
        Method to calculate the induced voltage at the current turn. DFTs are
        used for calculations in time and frequency domain (see classes below)
        """

        if self.n_fft not in beam_spectrum_dict:
            self.profile.beam_spectrum_generation(self.n_fft)
            beam_spectrum_dict[self.n_fft] = self.profile.beam_spectrum

        # self.profile.beam_spectrum_generation(self.n_fft)
        beam_spectrum = beam_spectrum_dict[self.n_fft]

        induced_voltage = - (self.beam.Particle.charge * e * self.beam.ratio
                             * bm.irfft(self.total_impedance.astype(dtype=bm.precision.complex_t, order='C', copy=False) * beam_spectrum))

        self.induced_voltage = induced_voltage[:self.n_induced_voltage].astype(
            dtype=bm.precision.real_t, order='C', copy=False)

    def induced_voltage_mtw(self, beam_spectrum_dict={}):
        """
        Method to calculate the induced voltage taking into account the effect
        from previous passages (multi-turn wake)
        """

        # Shift of the memory wake field by the current revolution period
        self.shift_trev()

        # Induced voltage of the current turn calculation
        self.induced_voltage_1turn(beam_spectrum_dict)

        # Setting to zero to the last part to remove the contribution from the
        # front wake
        self.induced_voltage[self.n_induced_voltage -
                             self.front_wake_buffer:] = 0

        # Add the induced voltage of the current turn to the memory from
        # previous turns
        self.mtw_memory[:self.n_induced_voltage] += self.induced_voltage

        self.induced_voltage = self.mtw_memory[:self.n_induced_voltage]

    def shift_trev_freq(self):
        """
        Method to shift the induced voltage by a revolution period in the
        frequency domain
        """

        t_rev = self.RFParams.t_rev[self.RFParams.counter[0]]
        # Shift in frequency domain
        induced_voltage_f = bm.rfft(self.mtw_memory, self.n_mtw_fft)
        induced_voltage_f *= bm.exp(self.omegaj_mtw * t_rev)
        self.mtw_memory = bm.irfft(induced_voltage_f)[:self.n_mtw_memory]
        # Setting to zero to the last part to remove the contribution from the
        # circular convolution
        self.mtw_memory[-int(self.buffer_size):] = 0

    def shift_trev_time(self):
        """
        Method to shift the induced voltage by a revolution period in the
        time domain (linear interpolation)
        """

        t_rev = self.RFParams.t_rev[self.RFParams.counter[0]]

        # self.mtw_memory = bm.interp_const_space(self.time_mtw + t_rev,
        self.mtw_memory = bm.interp(self.time_mtw + t_rev,
                                    self.time_mtw, self.mtw_memory,
                                    left=0, right=0)

    def _track(self):
        """
        Tracking method
        """

        self.induced_voltage_generation()

        bm.linear_interp_kick(dt=self.beam.dt, dE=self.beam.dE,
                              voltage=self.induced_voltage,
                              bin_centers=self.profile.bin_centers,
                              charge=self.beam.Particle.charge,
                              acceleration_kick=0.)


class InducedVoltageTime(_InducedVoltage):
    r"""
    Induced voltage derived from the sum of several wake fields (time domain)

    Parameters
    ----------
    Beam : object
        Beam object
    Profile : object
        Profile object
    wake_source_list : list
        Wake sources list (e.g. list of Resonator objects)
    wake_length : float, optional
        Wake length [s]
    multi_turn_wake : boolean, optional
        Multi-turn wake enable flag
    RFParams : object, optional
        RFStation object for turn counter and revolution period
    mtw_mode : boolean, optional
        Multi-turn wake mode can be 'freq' or 'time' (default)
    use_regular_fft : boolean
        use the next_regular function to ensure regular number for FFT
        calculations (default is True for efficient calculations, for
        better control of the sampling frequency False is preferred)

    Attributes
    ----------
    wake_source_list : list
        Wake sources list (e.g. list of Resonator objects)
    total_wake : float array
        Total wake array of all sources in :math:`\Omega / s`
    use_regular_fft : boolean
        User set value to use (default) or not regular numbers for FFTs
    """

    def __init__(self, Beam, Profile, wake_source_list, wake_length=None,
                 multi_turn_wake=False, RFParams=None, mtw_mode=None,
                 use_regular_fft=True):

        # Wake sources list (e.g. list of Resonator objects)
        self.wake_source_list = wake_source_list

        # Total wake array of all sources in :math:`\Omega / s`
        self.total_wake = 0

        # Call the __init__ method of the parent class [calls process()]
        _InducedVoltage.__init__(self, Beam, Profile, frequency_resolution=None,
                                 wake_length=wake_length, multi_turn_wake=multi_turn_wake,
                                 RFParams=RFParams, mtw_mode=mtw_mode,
                                 use_regular_fft=use_regular_fft)

    def process(self):
        """
        Reprocess the impedance contributions. To be run when profile changes
        """

        _InducedVoltage.process(self)

        # Number of points for the FFT, equal to the length of the induced
        # voltage array + number of profile -1 to calculate a linear convolution
        # in the frequency domain. The next regular number is used for speed,
        # therefore the frequency resolution is always equal or finer than
        # the input value
        if self.use_regular_fft:
            self.n_fft = next_regular(int(self.n_induced_voltage) +
                                      int(self.profile.n_slices) - 1)
        else:
            self.n_fft = int(self.n_induced_voltage) + \
                int(self.profile.n_slices) - 1

        # Frequency resolution in Hz
        self.frequency_resolution = 1 / (self.n_fft * self.profile.bin_size)

        # Time array of the wake in s
        self.time = np.arange(0, self.wake_length, self.wake_length
                              / self.n_induced_voltage,
                              dtype=bm.precision.real_t)

        # Processing the wakes
        self.sum_wakes(self.time)

    def sum_wakes(self, time_array):
        """
        Summing all the wake contributions in one total wake.
        """

        self.total_wake = np.zeros(time_array.shape)
        for wake_object in self.wake_source_list:
            wake_object.wake_calc(time_array)
            self.total_wake += wake_object.wake

        # Pseudo-impedance used to calculate linear convolution in the
        # frequency domain (padding zeros)
        self.total_impedance = bm.rfft(self.total_wake, self.n_fft)

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        import cupy as cp
        self.induced_voltage = cp.array(self.induced_voltage)
        self.time = cp.array(self.time)
        self.total_wake = cp.array(self.total_wake)
        self.total_impedance = cp.array(self.total_impedance)
        if hasattr(self, 'mtw_memory'):
            self.mtw_memory = cp.array(self.mtw_memory)
        if hasattr(self, 'time_mtw'):
            self.time_mtw = cp.array(self.time_mtw)
        if hasattr(self, 'omegaj_mtw'):
            self.omegaj_mtw = cp.array(self.omegaj_mtw)
        if hasattr(self, 'freq_mtw'):
            self.freq_mtw = cp.array(self.freq_mtw)
        if hasattr(self, 'total_wake'):
            self.total_wake = cp.array(self.total_wake)
        if hasattr(self, 'time'):
            self.time = cp.array(self.time)

        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        import cupy as cp
        self.induced_voltage = cp.asnumpy(self.induced_voltage)
        self.time = cp.asnumpy(self.time)
        self.total_wake = cp.asnumpy(self.total_wake)
        self.total_impedance = cp.asnumpy(self.total_impedance)
        if hasattr(self, 'mtw_memory'):
            self.mtw_memory = cp.asnumpy(self.mtw_memory)
        if hasattr(self, 'time_mtw'):
            self.time_mtw = cp.asnumpy(self.time_mtw)
        if hasattr(self, 'omegaj_mtw'):
            self.omegaj_mtw = cp.asnumpy(self.omegaj_mtw)
        if hasattr(self, 'freq_mtw'):
            self.freq_mtw = cp.asnumpy(self.freq_mtw)
        if hasattr(self, 'total_wake'):
            self.total_wake = cp.asnumpy(self.total_wake)
        if hasattr(self, 'time'):
            self.time = cp.asnumpy(self.time)

        # to make sure it will not be called again
        self._device = 'CPU'


class InducedVoltageFreq(_InducedVoltage):
    r"""
    Induced voltage derived from the sum of several impedances

    Parameters
    ----------
    Beam : object
        Beam object
    Profile : object
        Profile object
    impedance_source_list : list
        Impedance sources list (e.g. list of Resonator objects)
    frequency_resolution : float, optional
        Frequency resolution of the impedance [Hz]
    multi_turn_wake : boolean, optional
        Multi-turn wake enable flag
    front_wake_length : float, optional
        Lenght [s] of the front wake (if any) for multi-turn wake mode
    RFParams : object, optional
        RFStation object for turn counter and revolution period
    mtw_mode : boolean, optional
        Multi-turn wake mode can be 'freq' or 'time' (default)
    use_regular_fft : boolean
        use the next_regular function to ensure regular number for FFT
        calculations (default is True for efficient calculations, for
        better control of the sampling frequency False is preferred)

    Attributes
    ----------
    impedance_source_list : list
        Impedance sources list (e.g. list of Resonator objects)
    total_impedance : float array
        Total impedance array of all sources in* :math:`\Omega`
    front_wake_length : float
        Lenght [s] of the front wake (if any) for multi-turn wake mode
    use_regular_fft : boolean
        User set value to use (default) or not regular numbers for FFTs
    """

    def __init__(self, Beam, Profile, impedance_source_list,
                 frequency_resolution=None, multi_turn_wake=False,
                 front_wake_length=0, RFParams=None, mtw_mode=None,
                 use_regular_fft=True):

        # Impedance sources list (e.g. list of Resonator objects)
        self.impedance_source_list = impedance_source_list

        # Total impedance array of all sources in* :math:`\Omega`
        self.total_impedance = 0

        # Lenght in s of the front wake (if any) for multi-turn wake mode.
        # If the impedance calculation is performed in frequency domain, an
        # artificial front wake may appear. With this option, it is possible to
        # set to zero a portion at the end of the induced voltage array.*
        self.front_wake_length = front_wake_length

        # Call the __init__ method of the parent class
        _InducedVoltage.__init__(self, Beam, Profile, wake_length=None,
                                 frequency_resolution=frequency_resolution,
                                 multi_turn_wake=multi_turn_wake, RFParams=RFParams,
                                 mtw_mode=mtw_mode, use_regular_fft=use_regular_fft)

    def process(self):
        """
        Reprocess the impedance contributions. To be run when profile change
        """

        _InducedVoltage.process(self)

        # Number of points for the FFT. The next regular number is used for
        # speed, therefore the frequency resolution is always equal or finer
        # than the input value
        if self.use_regular_fft:
            self.n_fft = next_regular(self.n_induced_voltage)
        else:
            self.n_fft = self.n_induced_voltage

        self.profile.beam_spectrum_freq_generation(self.n_fft)

        # Frequency array and resolution of the impedance in Hz
        self.freq = self.profile.beam_spectrum_freq
        self.frequency_resolution = 1 / (self.n_fft * self.profile.bin_size)

        # Length of the front wake in frequency domain calculations
        if self.front_wake_length:
            self.front_wake_buffer = int(np.ceil(
                np.max(self.front_wake_length) / self.profile.bin_size))

        # Processing the impedances
        self.sum_impedances(self.freq)

    def sum_impedances(self, freq):
        """
        Summing all the wake contributions in one total impedance.
        """

        self.total_impedance = np.zeros(
            freq.shape, dtype=bm.precision.complex_t, order='C')

        for impedance_source in self.impedance_source_list:
            impedance_source.imped_calc(freq)
            self.total_impedance += impedance_source.impedance

        # Factor relating Fourier transform and DFT
        self.total_impedance /= self.profile.bin_size

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        import cupy as cp
        self.induced_voltage = cp.array(self.induced_voltage)
        self.freq = cp.array(self.freq)
        self.total_impedance = cp.array(self.total_impedance)
        if hasattr(self, 'mtw_memory'):
            self.mtw_memory = cp.array(self.mtw_memory)
        if hasattr(self, 'time_mtw'):
            self.time_mtw = cp.array(self.time_mtw)
        if hasattr(self, 'freq_mtw'):
            self.freq_mtw = cp.array(self.freq_mtw)
        if hasattr(self, 'omegaj_mtw'):
            self.omegaj_mtw = cp.array(self.omegaj_mtw)

        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        import cupy as cp
        self.induced_voltage = cp.asnumpy(self.induced_voltage)
        self.freq = cp.asnumpy(self.freq)
        self.total_impedance = cp.asnumpy(self.total_impedance)
        if hasattr(self, 'mtw_memory'):
            self.mtw_memory = cp.asnumpy(self.mtw_memory)
        if hasattr(self, 'time_mtw'):
            self.time_mtw = cp.asnumpy(self.time_mtw)
        if hasattr(self, 'freq_mtw'):
            self.freq_mtw = cp.asnumpy(self.freq_mtw)
        if hasattr(self, 'omegaj_mtw'):
            self.omegaj_mtw = cp.asnumpy(self.omegaj_mtw)

        # to make sure it will not be called again
        self._device = 'CPU'


class InductiveImpedance(_InducedVoltage):
    r"""
    Constant imaginary Z/n impedance

    Parameters
    ----------
    Beam : object
        Beam object
    Profile : object
        Profile object
    Z_over_n : float array
        Constant imaginary Z/n program in* :math:`\Omega`.
    RFParams : object
        RFStation object for turn counter and revolution period
    deriv_mode : string, optional
        Derivation method to compute induced voltage

    Attributes
    ----------
    Z_over_n : float array
        Constant imaginary Z/n program in* :math:`\Omega`.
    deriv_mode : string, optional
        Derivation method to compute induced voltage
    """

    def __init__(self, Beam, Profile, Z_over_n, RFParams,
                 deriv_mode='gradient'):

        # Constant imaginary Z/n program in* :math:`\Omega`.
        self.Z_over_n = Z_over_n

        # Derivation method to compute induced voltage
        self.deriv_mode = deriv_mode

        # Call the __init__ method of the parent class
        _InducedVoltage.__init__(self, Beam, Profile, RFParams=RFParams)

    def induced_voltage_1turn(self, beam_spectrum_dict={}):
        """
        Method to calculate the induced voltage through the derivative of the
        profile. The impedance must be a constant Z/n.
        """

        index = self.RFParams.counter[0]

        induced_voltage = - (self.beam.Particle.charge * e / (2 * np.pi) *
                             self.beam.ratio * self.Z_over_n[index] *
                             self.RFParams.t_rev[index] / self.profile.bin_size *
                             self.profile.beam_profile_derivative(self.deriv_mode)[1])

        self.induced_voltage = (induced_voltage[:self.n_induced_voltage]).astype(
            dtype=bm.precision.real_t, order='C', copy=False)

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        import cupy as cp
        self.induced_voltage = cp.array(self.induced_voltage)
        self.Z_over_n = cp.array(self.Z_over_n)
        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        import cupy as cp
        self.induced_voltage = cp.asnumpy(self.induced_voltage)
        self.Z_over_n = cp.asnumpy(self.Z_over_n)
        # to make sure it will not be called again
        self._device = 'CPU'


class InducedVoltageResonator(_InducedVoltage):
    r"""
    *Calculates the induced voltage of several resonators for arbitrary
    line density. It does so by linearily interpolating the line density and
    solving the convolution integral with the resonator impedance analytically.
    The line density need NOT be sampled at equidistant points. The times where
    the induced voltage is calculated need to be the same where the line
    density is sampled. If no timeArray is passed, the induced voltage is
    evaluated at the points of the line density. This is nececassry of
    compatability with other functions that calculate the induced voltage.
    Currently, it requires the all quality factors :math:`Q>0.5`
    Currently, only works for single turn.*

    Parameters
    ----------
    Beam : object
        Beam object
    Profile : object
        Profile object
    Resonators : object
        Resonators object
    timeArray : float array, optional
        Array of time values where the induced voltage is calculated.
        If left out, the induced voltage is calculated at the times of the line
        density.

    Attributes
    ----------
    beam : object
        Copy of the Beam object in order to access the beam info.
    profile : object
        Copy of the Profile object in order to access the line density.
    tArray : float array
        array of time values where the induced voltage is calculated.
        If left out, the induced voltage is calculated at the times of the
        line density
    atLineDensityTimes : boolean
        flag indicating if the induced voltage has to be computed for timeArray
        or for the line density
    n_time : int
        length of tArray
    R, omega_r, Q : lists of float
        Resonators parameters
    n_resonators : int
        Number of resonators
    induced_voltage : float array
        Computed induced voltage [V]
    """

    def __init__(self, Beam, Profile, Resonators, timeArray=None):

        # Test if one or more quality factors is smaller than 0.5.
        if sum(Resonators.Q < 0.5) > 0:
            # ResonatorError
            raise RuntimeError('All quality factors Q must be larger than 0.5')

        # Copy of the Beam object in order to access the beam info.
        self.beam = Beam
        # Copy of the Profile object in order to access the line density.
        self.profile = Profile

        # Optional array of time values where the induced voltage is calculated.
        # If left out, the induced voltage is calculated at the times of the
        # line density.
        if timeArray is None:
            self.tArray = self.profile.bin_centers
            self.atLineDensityTimes = True
        else:
            self.tArray = timeArray
            self.atLineDensityTimes = False

        # Length of timeArray
        self.n_time = len(self.tArray)

        # Copy of the shunt impedances of the Resonators in* :math:`\Omega`
        self.R = Resonators.R_S
        # Copy of the resonant frequencies of the Resonators in in 1/s
        self.omega_r = Resonators.omega_R  # resonant frequencies [1/s]
        # Copy of the quality factors of the Resonators
        self.Q = Resonators.Q
        # Number of resonators
        self.n_resonators = len(self.R)

        # For internal use
        self._Qtilde = self.Q * np.sqrt(1. - 1. / (4. * self.Q**2.))
        self._reOmegaP = self.omega_r * self._Qtilde / self.Q
        self._imOmegaP = self.omega_r / (2. * self.Q)

        # Each the 'n_resonator' rows of the matrix holds the induced voltage
        # at the 'n_time' time-values of one cavity. For internal use.
        self._tmp_matrix = np.ones(
            (self.n_resonators, self.n_time), dtype=bm.precision.real_t, order='C')

        # Slopes of the line segments. For internal use.
        self._kappa1 = np.zeros(
            int(self.profile.n_slices - 1), dtype=bm.precision.real_t, order='C')

        # Matrix to hold n_times many tArray[t]-bin_centers arrays.
        self._deltaT = np.zeros(
            (self.n_time, self.profile.n_slices), dtype=bm.precision.real_t, order='C')

        # Call the __init__ method of the parent class [calls process()]
        _InducedVoltage.__init__(self, Beam, Profile, wake_length=None,
                                 frequency_resolution=None,
                                 multi_turn_wake=False, RFParams=None, mtw_mode=None)

    def process(self):
        r"""
        Reprocess the impedance contributions. To be run when slicing changes
        """

        _InducedVoltage.process(self)

        # Since profile object changed, need to assign the proper dimensions to
        # _kappa1 and _deltaT
        self._kappa1 = np.zeros(
            int(self.profile.n_slices - 1), dtype=bm.precision.real_t, order='C')
        self._deltaT = np.zeros(
            (self.n_time, self.profile.n_slices), dtype=bm.precision.real_t, order='C')

    def induced_voltage_1turn(self, beam_spectrum_dict={}):
        r"""
        Method to calculate the induced voltage through linearily
        interpolating the line density and applying the analytic equation
        to the result.
        """

        # Compute the slopes of the line sections of the linearily interpolated
        # (normalized) line density.
        self._kappa1[:] = bm.diff(self.profile.n_macroparticles) \
            / bm.diff(self.profile.bin_centers) \
            / (self.beam.n_macroparticles * self.profile.bin_size)
        # [:] makes kappa pass by reference

        for t in range(self.n_time):
            self._deltaT[t] = self.tArray[t] - self.profile.bin_centers

        # For each cavity compute the induced voltage and store in the r-th row
        for r in range(self.n_resonators):
            tmp_sum = ((((2 *
                          bm.cos(self._reOmegaP[r] * self._deltaT)
                          + bm.sin(self._reOmegaP[r] * self._deltaT) / self._Qtilde[r]) *
                         bm.exp(-self._imOmegaP[r] * self._deltaT)) *
                        self.Heaviside(self._deltaT)) -
                       bm.sign(self._deltaT))
            # np.sum performs the sum over the points of the line density
            self._tmp_matrix[r] = self.R[r] / (2 * self.omega_r[r] * self.Q[r]) \
                * bm.sum(self._kappa1 * np.diff(tmp_sum), axis=1)

        # To obtain the voltage, sum the contribution of each cavity...
        self.induced_voltage = self._tmp_matrix.sum(axis=0)
        # ... and multiply with bunch charge
        self.induced_voltage *= -self.beam.Particle.charge * e \
            * self.beam.n_macroparticles * self.beam.ratio
        self.induced_voltage = self.induced_voltage.astype(
            dtype=bm.precision.real_t, order='C', copy=False)

    # Implementation of Heaviside function
    def Heaviside(self, x):
        r"""
        Heaviside function, which returns 1 if x>1, 0 if x<0, and 1/2 if x=0
        """
        return 0.5 * (bm.sign(x) + 1.)

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        import cupy as cp
        self.induced_voltage = cp.array(self.induced_voltage)
        self._kappa1 = cp.array(self._kappa1)
        self._deltaT = cp.array(self._deltaT)
        self.tArray = cp.array(self.tArray)
        self._tmp_matrix = cp.array(self._tmp_matrix)
        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        import cupy as cp
        self.induced_voltage = cp.asnumpy(self.induced_voltage)
        self._kappa1 = cp.asnumpy(self._kappa1)
        self._deltaT = cp.asnumpy(self._deltaT)
        self.tArray = cp.asnumpy(self.tArray)
        self._tmp_matrix = cp.asnumpy(self._tmp_matrix)

        # to make sure it will not be called again
        self._device = 'CPU'



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module gathering and processing all RF parameters used in the simulation.**

:Authors: **Alexandre Lasheen**, **Danilo Quartullo**, **Helga Timko**
'''

from __future__ import division, print_function

from builtins import range, str

import numpy as np
import scipy
from packaging.version import Version

if Version(scipy.__version__) >= Version("1.14"):
    from scipy.integrate import cumulative_trapezoid as cumtrapz
else:
    from scipy.integrate import cumtrapz


from scipy.constants import c

from ..beam.beam import Proton
from ..input_parameters.rf_parameters_options import RFStationOptions
from ..utils import bmath as bm


class RFStation:
    r""" Class containing all the RF parameters for all the RF systems in one
    ring segment or RF station.

    **How to use RF programs:**

    * For 1 RF system and constant values of V, h, or phi, input a single value
    * For 1 RF system and varying values of V, h, or phi, input an array of
      n_turns+1 values
    * For several RF systems and constant values of V, h, or phi, input lists
      of single values
    * For several RF systems and varying values of V, h, or phi, input lists
      of arrays of n_turns+1 values
    * For pre-processing, pass a list of times-voltages, times-harmonics,
      and/or times-phases for **each** RF system as a tuple
      ((time_1, voltage_1), (time_2, voltage_2), ...)

    Optional: RF frequency other than the design frequency. In this case, need
    to use a beam phase loop for correct RF phase!

    Optional: empty RFStation (e.g. for machines with synchrotron radiation);
    input voltage as 0.

    The index :math:`n` denotes time steps, :math:`l` the index of the RF
    systems in the section.

    **N.B. for negative eta the RF phase has to be shifted by Pi w.r.t the time
    reference.**

    Parameters
    ----------
    Ring : class
        A Ring type class
    harmonic : float (opt: float array/matrix, tuple of float array/matrix)
        Harmonic number of the RF system, :math:`h_{l,n}` [1]. For input
        options, see above
    voltage : float (opt: float array/matrix, tuple of float array/matrix)
        RF cavity voltage as seen by the beam, :math:`V_{l,n}` [V]. For input
        options, see above
    phi_rf_d : float (opt: float array/matrix, tuple of float array/matrix)
        Programmed/designed RF cavity phase,
        :math:`\phi_{d,l,n}` [rad]. For input options, see above
    n_rf : int
        Optional, Number of harmonic rf systems in the section :math:`l`.
        Becomes mandatory for several rf systems.
    section_index : int
        Optional, In case of several sections in the Ring object, this
        specifies after which section the rf station is located (to get the
        right momentum program etc.). Value should be in the range
        1..Ring.n_sections
    omega_rf : float (opt: float array/matrix)
        Optional, Sets the rf angular frequency program that does not follow
        the harmonic condition. For input options, see above.
    phi_noise : float (opt: float array/matrix)
        Optional, programmed RF cavity phase noise, :math:`\phi_{N,l,n}` [rad].
        Added to all RF systems in the station. For input options, see above
    phi_modulation : class (opt: iterable of classes)
        A PhaseModulation type class (or iterable of classes)
    RFStationOptions : class
        Optionnal, A RFStationOptions-based class defining smoothing,
        interpolation, etc. options for harmonic, voltage, and/or
        phi_rf_d programme to be interpolated to a turn-by-turn programme

    Attributes
    ----------
    counter : int
        Counter of the current simulation time step; defined as a list in
        order to be passed by reference
    section_index : int
        Unique index :math:`k` of the RF station the present class is defined
        for. Input in the range 1..n_sections (see
        :py:class:`input_parameters.ring.Ring`).
        Inside the code, indices 0..n_sections-1 are used.
    Particle : class
        Inherited from
        :py:attr:`input_parameters.ring.Ring.Particle`
    n_turns : int
        Inherited from
        :py:attr:`input_parameters.ring.Ring.n_turns`
    ring_circumference : float
        Inherited from
        :py:attr:`input_parameters.ring.Ring.ring_circumference`
    section_length : float
        Length :math:`L_k` of the RF section; inherited from
        :py:attr:`input_parameters.ring.Ring.ring_length`
    length_ratio : float
        Fractional RF section length :math:`L_k/C`
    t_rev : float array [n_turns+1]
        Inherited from
        :py:attr:`input_parameters.ring.Ring.t_rev`
    momentum : float array [n_turns+1]
        Momentum program of the present RF section; inherited from
        :py:attr:`input_parameters.ring.Ring.momentum`
    beta : float array [n_turns+1]
        Relativistic beta of the present RF section; inherited from
        :py:attr:`input_parameters.ring.Ring.beta`
    gamma : float array [n_turns+1]
        Relativistic gamma of the present RF section; inherited from
        :py:attr:`input_parameters.ring.Ring.gamma`
    energy : float array [n_turns+1]
        Total energy of the present RF section; inherited from
        :py:attr:`input_parameters.ring.Ring.energy`
    delta_E : float array [n_turns+1]
        Time derivative of total energy of the present section; inherited from
        :py:attr:`input_parameters.ring.Ring.delta_E`
    alpha_order : int
        Inherited from
        :py:attr:`input_parameters.ring.Ring.alpha_order`
    charge : int
        Inherited from
        :py:attr:`beam.Particle.charge`
    alpha_0 : float array [n_turns+1]
        Zeroth order momentum compaction factor of the present section;
        inherited from
        :py:attr:`input_parameters.ring.Ring.alpha_0`
    alpha_1 : float array [n_turns+1]
        First order momentum compaction factor of the present section;
        inherited from
        :py:attr:`input_parameters.ring.Ring.alpha_1`
    alpha_2 : float array [n_turns+1]
        Second order momentum compaction factor of the present section;
        inherited from
        :py:attr:`input_parameters.ring.Ring.alpha_2`
    eta_0 : float array [n_turns+1]
        Zeroth order slippage factor of the present section; inherited from
        :py:attr:`input_parameters.ring.Ring.eta_0`
    eta_1 : float array [n_turns+1]
        First order slippage factor of the present section; inherited from
        :py:attr:`input_parameters.ring.Ring.eta_1`
    eta_2 : float array [n_turns+1]
        Second order slippage factor of the present section; inherited from
        :py:attr:`input_parameters.ring.Ring.eta_2`
    sign_eta_0 : float array
        Sign of the eta_0 array
    harmonic : float matrix [n_rf, n_turns+1]
        Harmonic number for each rf system,
        :math:`h_{l,n}` [1]
    voltage : float matrix [n_rf, n_turns+1]
        Actual rf voltage of each harmonic system,
        :math:`V_{rf,l,n}` [V]
    empty : bool
        Flag to specify if the RFStation is empty
    phi_rf_d : float matrix [n_rf, n_turns+1]
        Designed rf cavity phase of each harmonic system,
        :math:`\phi_{d,l,n}` [rad]
    phi_rf : float matrix [n_rf, n_turns+1]
        Actual RF cavity phase of each harmonic system used for tracking,
        :math:`\phi_{rf,l,n}` [rad]. Initially the same as the designed phase.
    omega_rf_d : float matrix [n_rf, n_turns+1]
        Design RF angular frequency of the RF systems in the station
        :math:`\omega_{d,l,n} = \frac{h_{l,n} \beta_{l,n} c}{R_{s,n}}` [Hz]
    omega_rf : float matrix [n_rf, n_turns+1]
        Actual RF angular frequency of the RF systems in the station
        :math:`\omega_{rf,l,n} = \frac{h_{l,n} \beta_{l,n} c}{R_{s,n}}` [Hz].
        Initially the same as the designed angular frequency.
    phi_noise : None or float matrix [n_rf, n_turns+1]
        Programmed cavity phase noise for each RF harmonic.
    phi_modulation : None or float matrix [n_rf, n_turns+1]
        Programmed cavity phase modulation for each RF harmonic.
    dphi_rf : float matrix [n_rf]
        Accumulated RF phase error of each harmonic system
        :math:`\Delta \phi_{rf,l,n}` [rad]
    t_rf : float matrix [n_rf, n_turns+1]
        RF period :math:`\frac{2 \pi}{\omega_{rf,l,n}}` [s]
    phi_s : float array [n_turns+1]
        Synchronous phase for this section (if not empty), calculated in
        :py:func:`input_parameters.rf_parameters.calculate_phi_s`
    Q_s : float array [n_turns+1]
        Synchrotron tune for this section (if not empty), calculated in
        :py:func:`input_parameters.rf_parameters.calculate_Q_s`
    omega_s0 : float array [n_turns+1]
        Central synchronous angular frequency corresponding to Q_s (single
        harmonic, no intensity effects)
        :math:`\omega_{s,0} = Q_s \omega_{\text{rev}}` [1/s], where
        :math:`\omega_{\text{rev}}` is defined in
        :py:class:`input_parameters.ring.Ring`)
    RFStationOptions : RFStationOptions()
        The RFStationOptions is kept as an attribute of the RFStationg object
        for further usage.


    Examples
    --------
    >>> # To declare a double-harmonic RF system for protons:
    >>>
    >>> n_turns = 10
    >>> C = 26659
    >>> alpha = 3.21e-4
    >>> momentum = 450e9
    >>> ring = Ring(n_turns, C, alpha, momentum)
    >>> rf_station = RFStation(ring, [35640, 71280], [6e6, 6e5], [0, 0], 2)

    """

    def __init__(self, Ring, harmonic, voltage, phi_rf_d, n_rf=1,
                 section_index=1, omega_rf=None, phi_noise=None,
                 phi_modulation=None, RFStationOptions=RFStationOptions()):

        # Different indices
        self.counter = [int(0)]
        self.section_index = int(section_index - 1)
        if self.section_index < 0 \
                or self.section_index > Ring.n_sections - 1:
            raise RuntimeError("ERROR in RFStation: section_index out of" +
                               " allowed range!")
        self.n_rf = int(n_rf)

        # Imported from Ring
        self.Particle = Ring.Particle
        self.n_turns = Ring.n_turns
        self.ring_circumference = Ring.ring_circumference
        self.section_length = Ring.ring_length[self.section_index]
        self.length_ratio = float(self.section_length / self.ring_circumference)
        self.t_rev = Ring.t_rev
        self.momentum = Ring.momentum[self.section_index]
        self.beta = Ring.beta[self.section_index]
        self.gamma = Ring.gamma[self.section_index]
        self.energy = Ring.energy[self.section_index]
        self.delta_E = Ring.delta_E[self.section_index]
        self.alpha_order = Ring.alpha_order
        self.charge = self.Particle.charge

        # The order alpha_order used here can be replaced by Ring.alpha_order
        # when the assembler can differentiate the cases 'simple' and 'exact'
        # for the drift
        alpha_order = 2
        for i in range(alpha_order + 1):
            dummy = getattr(Ring, 'eta_' + str(i))
            setattr(self, "eta_%s" % i, dummy[self.section_index])
            dummy = getattr(Ring, 'alpha_' + str(i))
            setattr(self, "alpha_%s" % i, dummy[self.section_index])
        self.sign_eta_0 = np.sign(self.eta_0)

        # Reshape input rf programs
        # Reshape design harmonic
        self.harmonic = RFStationOptions.reshape_data(harmonic,
                                                      self.n_turns,
                                                      self.n_rf,
                                                      Ring.cycle_time,
                                                      Ring.RingOptions.t_start)
        self.harmonic = self.harmonic.astype(bm.precision.real_t, order='C', copy=False)

        # Reshape design voltage
        self.voltage = RFStationOptions.reshape_data(voltage,
                                                     self.n_turns,
                                                     self.n_rf,
                                                     Ring.cycle_time,
                                                     Ring.RingOptions.t_start)
        self.voltage = self.voltage.astype(bm.precision.real_t, order='C', copy=False)
        # Checking if the RFStation is empty
        if np.sum(self.voltage) == 0:
            self.empty = True
        else:
            self.empty = False

        # Reshape design phase
        self.phi_rf_d = RFStationOptions.reshape_data(phi_rf_d,
                                                      self.n_turns,
                                                      self.n_rf,
                                                      Ring.cycle_time,
                                                      Ring.RingOptions.t_start)

        # Calculating design rf angular frequency
        if omega_rf is None:
            self.omega_rf_d = 2. * np.pi * self.beta * c * self.harmonic / \
                (self.ring_circumference)
        else:
            self.omega_rf_d = RFStationOptions.reshape_data(
                omega_rf,
                self.n_turns,
                self.n_rf,
                Ring.cycle_time,
                Ring.RingOptions.t_start)
        self.omega_rf_d = self.omega_rf_d.astype(bm.precision.real_t, order='C', copy=False)

        # Reshape phase noise
        if phi_noise is not None:
            self.phi_noise = RFStationOptions.reshape_data(
                phi_noise,
                self.n_turns,
                self.n_rf,
                Ring.cycle_time,
                Ring.RingOptions.t_start)
            self.phi_noise = self.phi_noise.astype(bm.precision.real_t, order='C', copy=False)

        else:
            self.phi_noise = None

        if phi_modulation is not None:

            try:
                iter(phi_modulation)
            except TypeError:
                phi_modulation = [phi_modulation]

            dPhi = np.zeros([self.n_rf, self.n_turns + 1], dtype=bm.precision.real_t)
            dOmega = np.zeros([self.n_rf, self.n_turns + 1], dtype=bm.precision.real_t)
            for pMod in phi_modulation:
                system = np.where(self.harmonic[:, 0] == pMod.harmonic)[0]
                if len(system) == 0:
                    raise ValueError("No matching harmonic in phi_modulation")
                elif len(system) > 1:
                    raise RuntimeError("""Phase modulation not yet 
                                       implemented with multiple systems 
                                       at the same harmonic.""")
                else:
                    system = system[0]

                pMod.calc_modulation()
                pMod.calc_delta_omega((Ring.cycle_time, self.omega_rf_d[system]))
                dPhiInput, dOmegaInput = pMod.extend_to_n_rf(self.harmonic[:, 0])
                dPhi += RFStationOptions.reshape_data(dPhiInput,
                                                      self.n_turns,
                                                      self.n_rf,
                                                      Ring.cycle_time,
                                                      Ring.RingOptions.t_start)
                dOmega += RFStationOptions.reshape_data(dOmegaInput,
                                                        self.n_turns,
                                                        self.n_rf,
                                                        Ring.cycle_time,
                                                        Ring.RingOptions.t_start)

            self.phi_modulation = (dPhi, dOmega)
        else:

            self.phi_modulation = None

        # Copy of the desing rf programs in the one used for tracking
        # and that can be changed by feedbacks
        self.phi_rf = np.array(self.phi_rf_d).astype(bm.precision.real_t)
        self.dphi_rf = np.zeros(self.n_rf).astype(bm.precision.real_t)
        self.omega_rf = np.array(self.omega_rf_d).astype(bm.precision.real_t)
        self.t_rf = 2 * np.pi / self.omega_rf

        # From helper functions
        if not self.empty:
            self.phi_s = calculate_phi_s(self, self.Particle)
            self.Q_s = calculate_Q_s(self, self.Particle)
            self.omega_s0 = self.Q_s * Ring.omega_rev

    def eta_tracking(self, beam, counter, dE):
        r"""Function to calculate the slippage factor as a function of the
        energy offset :math:`\Delta E` of the particle. The slippage factor
        of the :math:`i` th order is :math:`\eta(\delta) = \sum_{i}(\eta_i \,
        \delta^i) = \sum_{i} \left(\eta_i \, \left[ \frac{\Delta E}
        {\beta_s^2 E_s} \right]^i \right)`

        """

        if self.alpha_order == 0:
            return self.eta_0[counter]
        else:
            eta = 0
            delta = dE / (beam.beta**2 * beam.energy)
            for i in range(self.alpha_order + 1):
                eta_i = getattr(self, 'eta_' + str(i))[counter]
                eta += eta_i * (delta**i)
            return eta

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return
        import cupy as cp
        if self.phi_modulation is not None:
            self.phi_modulation = (cp.array(self.phi_modulation[0]),
                                   cp.array(self.phi_modulation[1]))
        if self.phi_noise is not None:
            self.phi_noise = cp.array(self.phi_noise)

        self.voltage = cp.array(self.voltage)
        self.omega_rf = cp.array(self.omega_rf)
        self.phi_rf = cp.array(self.phi_rf)
        self.omega_rf_d = cp.array(self.omega_rf_d)
        self.harmonic = cp.array(self.harmonic)
        self.dphi_rf = cp.array(self.dphi_rf)
        self.t_rf = cp.array(self.t_rf)
        self.t_rev = cp.array(self.t_rev)

        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return
        import cupy as cp
        if self.phi_modulation is not None:
            self.phi_modulation = (cp.asnumpy(self.phi_modulation[0]),
                                   cp.asnumpy(self.phi_modulation[1]))
        if self.phi_noise is not None:
            self.phi_noise = cp.asnumpy(self.phi_noise)

        self.voltage = cp.asnumpy(self.voltage)
        self.omega_rf = cp.asnumpy(self.omega_rf)
        self.phi_rf = cp.asnumpy(self.phi_rf)
        self.omega_rf_d = cp.asnumpy(self.omega_rf_d)
        self.harmonic = cp.asnumpy(self.harmonic)
        self.dphi_rf = cp.asnumpy(self.dphi_rf)
        self.t_rf = cp.asnumpy(self.t_rf)
        self.t_rev = cp.asnumpy(self.t_rev)
        # to make sure it will not be called again
        self._device = 'CPU'


def calculate_Q_s(RFStation, Particle=Proton()):
    r""" Function calculating the turn-by-turn synchrotron tune for
    single-harmonic RF, without intensity effects.

    Parameters
    ----------
    RFStation : class
        An RFStation type class.
    Particle : class
        A Particle type class; default is Proton().

    Returns
    -------
    float
        Synchrotron tune.

    """

    return np.sqrt(RFStation.harmonic[0] * np.abs(Particle.charge) *
                   RFStation.voltage[0] *
                   np.abs(RFStation.eta_0 * np.cos(RFStation.phi_s)) /
                   (2 * np.pi * RFStation.beta**2 * RFStation.energy))


def calculate_phi_s(RFStation, Particle=Proton(),
                    accelerating_systems='as_single'):
    r"""Function calculating the turn-by-turn synchronous phase according to
    the parameters in the RFStation object. The phase is expressed in
    the lowest RF harmonic and with respect to the RF bucket (see the equations
    of motion defined for BLonD). The returned value is given in the range [0,
    2*Pi]. Below transition, the RF wave is shifted by Pi w.r.t. the time
    reference.

    The accelerating_systems option can be set to

    * 'as_single' (default): the synchronous phase is calculated analytically
      taking into account the phase program (RFStation.phi_offset).
    * 'all': the synchronous phase is calculated numerically by finding the
      minimum of the potential well; no intensity effects included. In case of
      several minima, the deepest is taken. **WARNING:** in case of RF
      harmonics with comparable voltages, this may lead to inconsistent
      values of phi_s.
    * 'first': not yet implemented. Its purpose should be to adjust the
      RFStation.phi_offset of the higher harmonics so that only the
      main harmonic is accelerating.

    Parameters
    ----------
    RFStation : class
        An RFStation type class.
    Particle : class
        A Particle type class; default is Proton().
    accelerating_systems : str
        Choice of accelerating systems; or options, see list above.

    Returns
    -------
    float
        Synchronous phase.

    """

    eta0 = RFStation.eta_0

    if accelerating_systems == 'as_single':

        denergy = np.append(RFStation.delta_E, RFStation.delta_E[-1])
        acceleration_ratio = denergy / (Particle.charge * RFStation.voltage[0, :])
        acceleration_test = ((acceleration_ratio > -1) & (acceleration_ratio < 1)) == 0
        
        # Validity check on acceleration_ratio
        if np.count_nonzero(acceleration_test) > 0:
            print("WARNING in calculate_phi_s(): acceleration is not " +
                  "possible (momentum increment is too big or voltage too " +
                  "low) at index " + str(acceleration_test.nonzero()[0]))

        phi_s = np.arcsin(acceleration_ratio)

        # Identify where eta swaps sign
        eta0_middle_points = (eta0[1:] + eta0[:-1]) / 2
        eta0_middle_points = np.append(eta0_middle_points, eta0[-1])
        index = np.where(eta0_middle_points > 0)[0]
        index_below = np.where(eta0_middle_points < 0)[0]

        # Project phi_s in correct range
        phi_s[index] = (np.heaviside(np.sign(Particle.charge), 0) * np.pi - phi_s[index]) % (2 * np.pi)
        phi_s[index_below] = (np.heaviside(np.sign(Particle.charge), 0) * np.pi + phi_s[index_below])\
            % (2 * np.pi)
        # phi_s[index] = (np.pi - phi_s[index]) % (2*np.pi)
        # phi_s[index_below] = (np.pi + phi_s[index_below]) % (2*np.pi)

        return phi_s

    elif accelerating_systems == 'all':

        phi_s = np.zeros(len(RFStation.voltage[0, 1:]))

        for indexTurn in range(len(RFStation.delta_E)):

            totalRF = 0
            if np.sign(eta0[indexTurn]) > 0:
                phase_array = np.linspace(
                    -float(RFStation.phi_rf[0, indexTurn + 1]),
                    -float(RFStation.phi_rf[0, indexTurn + 1]) + 2 * np.pi, 1000)
            else:
                phase_array = np.linspace(
                    -float(RFStation.phi_rf[0, indexTurn + 1]) - np.pi,
                    -float(RFStation.phi_rf[0, indexTurn + 1]) + np.pi, 1000)

            for indexRF in range(len(RFStation.voltage[:, indexTurn + 1])):
                totalRF += RFStation.voltage[indexRF, indexTurn + 1] * \
                    np.sin(RFStation.harmonic[indexRF, indexTurn + 1] /
                           np.min(RFStation.harmonic[:, indexTurn + 1]) *
                           phase_array +
                           RFStation.phi_rf[indexRF, indexTurn + 1])

            potential_well = - cumtrapz(
                np.sign(eta0[indexTurn]) * (totalRF -
                                            RFStation.delta_E[indexTurn] /
                                            abs(Particle.charge)),
                dx=phase_array[1] - phase_array[0], initial=0)

            phi_s[indexTurn] = np.mean(phase_array[
                potential_well == np.min(potential_well)])

        phi_s = np.insert(phi_s, 0, phi_s[0]) + RFStation.phi_rf[0, :]
        phi_s[eta0 < 0] += np.pi
        phi_s = phi_s % (2 * np.pi)

        return phi_s

    elif accelerating_systems == 'first':

        print("WARNING in calculate_phi_s(): accelerating_systems 'first'" +
              " not yet implemented")
        pass
    else:
        raise RuntimeError("ERROR in calculate_phi_s(): unrecognised" +
                           " accelerating_systems option")



'''
Basic methods and objects related to the GPU computational core.

@author: Konstantinos Iliakis
@date: 25.05.2023
'''

import os

from ..utils import precision


class GPUDev:
    """GPU Device object, singleton class

    Returns:
        _type_: _description_
    """
    __instance = None

    def __init__(self):
        """Initializes the GPU device object.

        Args:
            _gpu_num (int, optional): The device id. Defaults to 0.
        """
        if GPUDev.__instance is not None:
            return
        GPUDev.__instance = self
        self.id = -1
        self.dev = None
        self.name = ''
        self.attributes = {}
        self.attributes = {}
        self.grid_size = 0
        self.block_size = 0
        self.mod = None

    def set(self, _gpu_num=0):
        import cupy as cp

        self.id = _gpu_num
        self.dev = cp.cuda.Device(self.id)
        self.dev.use()

        self.name = cp.cuda.runtime.getDeviceProperties(self.dev)['name']
        self.attributes = self.dev.attributes
        self.properties = cp.cuda.runtime.getDeviceProperties(self.dev)

        # set the default grid and block sizes
        default_blocks = 2 * self.attributes['MultiProcessorCount']
        default_threads = self.attributes['MaxThreadsPerBlock']
        blocks = int(os.environ.get('GPU_BLOCKS', default_blocks))
        threads = int(os.environ.get('GPU_THREADS', default_threads))
        self.grid_size = (blocks, 1, 1)
        self.block_size = (threads, 1, 1)

        self.mod = None
        self.load_library(precision.str)

    def report_attributes(self):
        """ Stores in file device attributes
        """
        # Saves into a file all the device attributes
        with open(f'{self.name}-attributes.txt', 'w') as file:
            for key, value in self.attributes.items():
                file.write(f"{key}:{value}\n")

    def func(self, name):
        """Get kernel from kernel module

        Args:
            name (string): Kernel name

        Returns:
            _type_: _description_
        """
        return self.mod.get_function(name)
    
    def load_library(self, _precision):
        '''Load the GPU library

        Args:
            _precision (str): must be either 'single' or 'double'
        '''
        import cupy as cp
        assert _precision in ['single', 'double']

        this_dir = os.path.dirname(os.path.realpath(__file__)) + "/"
        comp_capability = self.dev.compute_capability
        library_path = os.path.join(this_dir, f'../gpu/kernels_sm_{comp_capability}_{_precision}.cubin')
        if os.path.exists(library_path):
            self.mod = cp.RawModule(path=library_path)


# Initialize empty GPU object
GPU_DEV = GPUDev()




# Copyright 2015 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Methods to generate RF phase modulation from given frequency, amplitude
and offset functions**

:Authors: **Simon Albright**
'''

# General imports
import numpy as np

# BLonD imports
import blond.utils.data_check as dCheck


class PhaseModulation:

    def __init__(self, timebase, frequency, amplitude, offset,
                 harmonic, multiplier=1, modulate_frequency=True):

        msg = "must be a single numerical value or have shape (2, n)"
        dCheck.check_input(timebase, "Timebase must have shape (n)", [-1])
        dCheck.check_input(frequency, "Frequency " + msg, 0, (2, -1))
        dCheck.check_input(amplitude, "Amplitude " + msg, 0, (2, -1))
        dCheck.check_input(offset, "Offset " + msg, 0, (2, -1))
        dCheck.check_input(multiplier, "Multiplier " + msg, 0, (2, -1))
        dCheck.check_input(
            harmonic, "Harmonic must be single valued number", 0)

        self.timebase = timebase
        self.frequency = frequency
        self.amplitude = amplitude
        self.offset = offset
        self.multiplier = multiplier
        self.harmonic = harmonic

        if not isinstance(modulate_frequency, bool):
            raise TypeError("modulate_frequency must be boolean")

        self._mod_freq = modulate_frequency

    # Calculate the modulation with linear interpolation of functions
    def calc_modulation(self):

        amplitude = self._interp_param(self.amplitude)
        frequency = self._interp_param(self.frequency)
        offset = self._interp_param(self.offset)
        multiplier = self._interp_param(self.multiplier)

        frequency *= multiplier

        self.dphi = amplitude \
            * np.sin(2 * np.pi * (np.cumsum(frequency
                                            * np.gradient(self.timebase))))\
            + offset

    def calc_delta_omega(self, omegaProg):

        dCheck.check_input(omegaProg, "omegaProg must have shape (2, n)",
                           (2, -1))

        if not self._mod_freq:
            self.domega = np.zeros(len(self.dphi))

        else:
            omega = self._interp_param(omegaProg)
            self.domega = np.gradient(self.dphi) * omega \
                / (2 * np.pi * self.harmonic)

    # Interpolate functions onto self.timebase
    def _interp_param(self, param):

        if dCheck.check_data_dimensions(param, 0)[0]:
            return np.array([param] * len(self.timebase))

        elif dCheck.check_data_dimensions(param, (2, -1))[0]:
            return np.interp(self.timebase, param[0], param[1])

        else:
            raise TypeError("Param must be number or have shape (2, n)")

    # Extend passed parameter to requred n_rf if n_rf > 1 for treatment in
    # rf_parameters
    def extend_to_n_rf(self, harmonics):

        try:
            n_rf = len(harmonics)
        except TypeError:
            n_rf = 1
            harmonics = [harmonics]

        if self.harmonic not in harmonics:
            raise AttributeError("self.harmonic not in harmonics")

        if not hasattr(self, 'domega'):
            raise AttributeError("""domega has not yet been calculated, 
                                 calc_delta_omega must be called first""")

        if n_rf == 1:
            return (self.timebase, self.dphi), (self.timebase, self.domega)

        else:
            extendTuple = ([self.timebase[0], self.timebase[-1]], [0, 0])
            return (tuple([self.timebase, self.dphi]
                          if self.harmonic == harmonics[i]
                          else extendTuple for i in range(n_rf)),

                    tuple([self.timebase, self.domega]
                          if self.harmonic == harmonics[i]
                          else extendTuple for i in range(n_rf)))



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Transfer function analysis for open and closed loop**

:Authors: **Jelena Banjac, Helga Timko**
'''

from __future__ import division
import numpy as np
import numpy.fft as npfft
from matplotlib.mlab import psd, csd
import matplotlib.pyplot as plt
import scipy.signal as scs

# Set up logging
import logging
logger = logging.getLogger(__name__)

class TransferFunction(object):
    r'''Reconstructing the transfer function of a DUT based on input and output
    signals.

    Parameters
    ----------
    signal_in : float array
        Complex signal going into the DUT
    signal_out : float array
        Complex signal coming out from the DUT
    T_s : float
        sampling time
    plot : bool
        Enable/disable plotting; default is False
    '''

    def __init__(self, signal_in, signal_out, T_s, plot=False):

        # Set up logging
        self.logger = logging.getLogger(__class__.__name__)
        self.logger.info("TransferFunction class initialized")

        # Read in variables
        self.signal_in = signal_in
        self.signal_out = signal_out
        self.T_s = T_s
        self.plot =plot


    def analyse(self, data_cut):

        # input signal
        input_signal = self.signal_in[data_cut:]
        self.logger.debug(f"OUTPUT signal: {input_signal.shape}")

        # input signal
        output_signal = self.signal_out[data_cut:]
        self.logger.debug(f"INPUT signal: {output_signal.shape}")

        # Estimate transfer function
        self.f_est, self.H_est = self.estimate_transfer_function(input_signal,
            output_signal, self.T_s, self.plot, self.logger)


    @staticmethod
    def estimate_transfer_function(input, output, T_s, plot, logger):
        # Calculate transfer function
        f_s = 1/T_s
        n_fft = int(np.floor(len(input)/4))
        f_est, H_est = tf_estimate(input, output, window=np.hamming(n_fft),
                                   noverlap=0, Fs=f_s, NFFT=n_fft)
        if plot:
            TransferFunction.plot_magnitude_and_phase(f_est, H_est)

        # reorder results to be form -freq to +freq
        f_est = [x + f_s/2 if x < 0 else x - f_s/2 for x in f_est]

        # Spectrum of the input signal
        f_max = TransferFunction.input_signal_spectrum(input, n_fft, f_s, plot)
        f_baseband = npfft.fftshift(f_est)

        low = np.where((0 > f_baseband) & (f_baseband > -f_max))[0]
        high = np.where((0 < f_baseband) & (f_baseband < f_max))[0]

        # Reconstruct the transfer function
        f = np.concatenate([f_baseband[low], f_baseband[high]])
        H = np.concatenate([H_est[low], H_est[high]])

        logger.debug("Transfer function reconstructed")

        if plot:
            TransferFunction.plot_magnitude_and_phase(f, H)

        return f, H


    @staticmethod
    def input_signal_spectrum(input, n, f_s, plot):
        """
        Prepare for drawing the input signal spectrum.

        Parameters
        ----------
        input : complex array
            Input signal of the system
        n : integer
            Number of points in the spectrum
        f_s : float
            Sampling frequency [1/s]

        Returns
        -------
        f_max : double
            Maximum frequency

        Attributes
        ----------
        N_sp: float
            Number of sections that the N-data spectrum is divided to
            calculate periodogram [1..N]
        f_m : ndarray
            Array of sample frequencies.
        P_ss: ndarray
            Power spectral density or power spectrum of input signal `In`.
            Afterwards, it shifted the
            zero-frequency component to the center of the spectrum.
        P_min : float
            -Pmin dB defines the minimum signal [35 dB].
        P_max : float
            Defines the maximum signal, depending on a `P_ss`
        P_bw : float
            Signal bandwidth.
        interval : int array
            Interval in which the P_ss signal is greater that min P_ss
        f_m_bw :
            Array of sample frequencies for a bandwidth of power spectrum.
        P_ss_bw :
            Bandwidth of power spectral density or power spectrum of input
            signal `In`.
        """
        N_sp = 16

        # power spectral density or power spectrum of input signal `input`.
        f_m, P_ss = scs.welch(input, window='hamming', noverlap=0,
                              nperseg=int(np.floor(n / N_sp)), fs=f_s)

        # reorder results to be form -freq to +freq
        f_m = [f - f_s if f >= f_s / 2 else f for f in f_m]

        # shift zero-frequency component to the center of the spectrum
        f_m, P_ss = npfft.fftshift(f_m).T, npfft.fftshift(P_ss)

        # minimum signal and maximum signal
        P_min, P_max = 35, 20 * np.log10(np.max(np.abs(P_ss)))

        # signal bandwidth
        P_bw = P_max - P_min

        # interval in which the P_ss signal is greater that min P_ss
        interval = np.where(abs(P_ss) >= 10 ** (P_bw / 20))[0]

        f_m_bw, P_ss_bw = f_m[interval], P_ss[interval]

        f_max = min(-min(f_m_bw), max(f_m_bw))

        return f_max

    @staticmethod
    def plot_magnitude_and_phase(freq, H):

        magnitude = 20 * np.log10(np.abs(H))
        phases = np.unwrap(np.angle(H))

        plt.rc('axes', labelsize=12, labelweight='normal', axisbelow=True)
        plt.rc('legend', fontsize=11, fancybox=True, framealpha=0.5)
        plt.rc('figure', figsize=(8, 6), titlesize=14)
        plt.rc('grid', c='lightgray')

        fig = plt.figure()
        gs = plt.GridSpec(2, 1)
        ax1 = fig.add_subplot(gs[0, 0])
        ax1.set_title('Transfer function')
        ax1.plot(freq / 10**6, magnitude, 'r', linewidth=0.3)
        ax1.set_xlabel('Frequency [MHz]')
        ax1.set_ylabel('Gain [dB]')

        ax2 = fig.add_subplot(gs[1, 0], sharex=ax1)
        ax2.plot(freq / 10**6, (180/np.pi)*phases, 'r', linewidth=0.3)
        ax2.set_xlabel('Frequency [MHz]')
        ax2.set_ylabel('Phase [degrees]')
        plt.show()


def tf_estimate(x, y, *args, **kwargs):
    """
    Estimate transfer function from x to y, see csd (from
    matplotlib.mlab package) for calling convention.
    Link: https://stackoverflow.com/questions/28462144/python-version-of
    -matlab-signal-toolboxs-tfestimate

    The vectors *x* and *y* are divided into *NFFT* length segments.
    Each segment is detrended by function *detrend* and windowed by
    function *window*.
    *noverlap* gives the length of the overlap between segments.

    Parameters
    ----------
    x : 1-D arrays or sequences
        Arrays or sequences containing the data
    y : 1-D arrays or sequences
        Arrays or sequences containing the data
    args : Default keyword values: None
    kwargs : NFFT, Fs, detrend, window, noverlap, pad_to, sides, scale_by_freq

    Returns
    -------
    frequency : float array
        Frequency array
    H : complex array
        Transfer function estimate

    Attributes
    ----------
    p_xy: float array
        Cross spectral density
    p_xx: float array
        Power spectral density
    frequencies_csd: array
        Frequency array for cross spectral density
    frequencies_psd: array
        Frequency array for power spectral density
    """
    p_xy, frequencies_csd = csd(y, x, *args, **kwargs)
    p_xx, frequencies_psd = psd(x, *args, **kwargs)

    return frequencies_csd, (p_xy/p_xx).conjugate()





# coding: utf-8
# Copyright 2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to compute the beam profile through slices**

:Authors: **Danilo Quartullo**, **Alexandre Lasheen**,
          **Juan F. Esteban Mueller**
'''

from __future__ import division, print_function

import numpy as np
# from numpy.fft import rfft, rfftfreq
from scipy import ndimage

from ..toolbox import filters_and_fitting as ffroutines
from ..utils import bmath as bm


class CutOptions:
    r"""
    This class groups all the parameters necessary to slice the phase space
    distribution according to the time axis, apart from the array collecting
    the profile which is defined in the constructor of the class Profile below.

    Parameters
    ----------
    cut_left : float
        Left edge of the slicing (optional). A default value will be set if
        no value is given.
    cut_right : float
        Right edge of the slicing (optional). A default value will be set
        if no value is given.
    n_slices : int
        Optional input parameters, corresponding to the number of
        :math:`\sigma_{RMS}` of the Beam to slice (this will overwrite
        any input of cut_left and cut_right).
    n_sigma : float
        defines the left and right extremes of the profile in case those are
        not given explicitly
    cuts_unit : str
        the unit of cut_left and cut_right, it can be seconds 's' or radians
        'rad'
    RFSectionParameters : object
        RFSectionParameters[0][0] is necessary for the conversion from radians
        to seconds if cuts_unit = 'rad'. RFSectionParameters[0][0] is the value
        of omega_rf of the main harmonic at turn number 0

    Attributes
    ----------
    cut_left : float
    cut_right : float
    n_slices : int
    n_sigma : float
    cuts_unit : str
    RFSectionParameters : object
    edges : float array
        contains the edges of the slices
    bin_centers : float array
        contains the centres of the slices

    Examples
    --------
    >>> from input_parameters.ring import Ring
    >>> from input_parameters.rf_parameters import RFStation
    >>> self.ring = Ring(n_turns = 1, ring_length = 100,
    >>> alpha = 0.00001, momentum = 1e9)
    >>> self.rf_params = RFStation(Ring=self.ring, n_rf=1, harmonic=[4620],
    >>>                  voltage=[7e6], phi_rf_d=[0.])
    >>> CutOptions = profileModule.CutOptions(cut_left=0, cut_right=2*np.pi,
    >>> n_slices = 100, cuts_unit='rad', RFSectionParameters=self.rf_params)

    """

    def __init__(self, cut_left=None, cut_right=None, n_slices=100,
                 n_sigma=None, cuts_unit='s', RFSectionParameters=None):
        """
        Constructor
        """

        if cut_left is not None:
            self.cut_left = float(cut_left)
        else:
            self.cut_left = cut_left

        if cut_right is not None:
            self.cut_right = float(cut_right)
        else:
            self.cut_right = cut_right

        self.n_slices = int(n_slices)

        if n_sigma is not None:
            self.n_sigma = float(n_sigma)
        else:
            self.n_sigma = n_sigma

        self.cuts_unit = str(cuts_unit)

        self.RFParams = RFSectionParameters

        if self.cuts_unit == 'rad' and self.RFParams is None:
            # CutError
            raise RuntimeError('You should pass an RFParams object to ' +
                               'convert from radians to seconds')
        if self.cuts_unit not in ['rad', 's']:
            # CutError
            raise RuntimeError('cuts_unit should be "s" or "rad"')

        self.edges = np.zeros(n_slices + 1, dtype=bm.precision.real_t, order='C')
        self.bin_centers = np.zeros(n_slices, dtype=bm.precision.real_t, order='C')
        self.bin_size = 0.0
        # For CuPy backend
        self._device = 'CPU'

    def set_cuts(self, Beam=None):
        r"""
        Method to set self.cut_left, self.cut_right, self.edges and
        self.bin_centers attributes.
        The frame is defined by :math:`n\sigma_{RMS}` or manually by the user.
        If not, a default frame consisting of taking the whole bunch +5% of the
        maximum distance between two particles in the bunch will be taken
        in each side of the frame.
        """

        if self.cut_left is None and self.cut_right is None:

            if self.n_sigma is None:
                dt_min = Beam.dt.min()
                dt_max = Beam.dt.max()
                self.cut_left = dt_min - 0.05 * (dt_max - dt_min)
                self.cut_right = dt_max + 0.05 * (dt_max - dt_min)
            else:
                mean_coords = np.mean(Beam.dt)
                sigma_coords = np.std(Beam.dt)
                self.cut_left = mean_coords - self.n_sigma * sigma_coords / 2
                self.cut_right = mean_coords + self.n_sigma * sigma_coords / 2

        else:

            self.cut_left = float(self.convert_coordinates(self.cut_left,
                                                           self.cuts_unit))
            self.cut_right = float(self.convert_coordinates(self.cut_right,
                                                            self.cuts_unit))

        self.edges = np.linspace(self.cut_left, self.cut_right,
                                 self.n_slices + 1).astype(dtype=bm.precision.real_t, order='C', copy=False)
        self.bin_centers = (self.edges[:-1] + self.edges[1:]) / 2
        self.bin_size = (self.cut_right - self.cut_left) / self.n_slices

    def track_cuts(self, Beam):
        """
        Track the slice frame (limits and slice position) as the mean of the
        bunch moves.
        Requires Beam statistics!
        Method to be refined!
        """

        delta = Beam.mean_dt - 0.5 * (self.cut_left + self.cut_right)

        self.cut_left += delta
        self.cut_right += delta
        self.edges += delta
        self.bin_centers += delta

    def convert_coordinates(self, value, input_unit_type):
        """
        Method to convert a value from 'rad' to 's'.
        """

        if input_unit_type == 's':
            return value

        elif input_unit_type == 'rad':
            return value /\
                self.RFParams.omega_rf[0, self.RFParams.counter[0]]

    def get_slices_parameters(self):
        """
        Reuturn all the computed parameters.
        """
        return self.n_slices, self.cut_left, self.cut_right, self.n_sigma, \
            self.edges, self.bin_centers, self.bin_size

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        # transfer recursively objects
        if recursive and self.RFParams:
            self.RFParams.to_gpu()

        import cupy as cp

        self.edges = cp.array(self.edges)
        self.bin_centers = cp.array(self.bin_centers)

        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        # transfer recursively objects
        if recursive and self.RFParams:
            self.RFParams.to_cpu()

        import cupy as cp

        self.edges = cp.asnumpy(self.edges)
        self.bin_centers = cp.asnumpy(self.bin_centers)

        if hasattr(self, 'rf_voltage'):
            self.rf_voltage = cp.asnumpy(self.rf_voltage)

        # to make sure it will not be called again
        self._device = 'CPU'


class FitOptions:
    """
    This class defines the method to be used turn after turn to obtain the
    position and length of the bunch profile.

    Parameters
    ----------

    fit_method : string
        Current options are 'gaussian',
        'fwhm' (full-width-half-maximum converted to 4 sigma gaussian bunch)
        and 'rms'. The methods 'gaussian' and 'rms' give both 4 sigma.
    fitExtraOptions : unknown
        For the moment no options can be passed into fitExtraOptions

    Attributes
    ----------

    fit_method : string
    fitExtraOptions : unknown
    """

    def __init__(self, fit_option=None, fitExtraOptions=None):
        """
        Constructor
        """

        self.fit_option = str(fit_option)
        self.fitExtraOptions = fitExtraOptions


class FilterOptions:

    """
    This class defines the filter to be used turn after turn to smooth
    the bunch profile.

    Parameters
    ----------

    filterMethod : string
        The only option available is 'chebishev'
    filterExtraOptions : dictionary
        Parameters for the Chebishev filter (see the method
        beam_profile_filter_chebyshev in filters_and_fitting.py in the toolbox
        package)

    Attributes
    ----------

    filterMethod : string
    filterExtraOptions : dictionary

    """

    def __init__(self, filterMethod=None, filterExtraOptions=None):
        """
        Constructor
        """

        self.filterMethod = str(filterMethod)
        self.filterExtraOptions = filterExtraOptions


class OtherSlicesOptions:

    """
    This class groups all the remaining options for the Profile class.

    Parameters
    ----------

    smooth : boolean
        If set True, this method slices the bunch not in the
        standard way (fixed one slice all the macroparticles contribute
        with +1 or 0 depending if they are inside or not). The method assigns
        to each macroparticle a real value between 0 and +1 depending on its
        time coordinate. This method can be considered a filter able to smooth
        the profile.
    direct_slicing : boolean
        If set True, the profile is calculated when the Profile class below
        is created. If False the user has to manually track the Profile object
        in the main file after its creation

    Attributes
    ----------

    smooth : boolean
    direct_slicing : boolean

    """

    def __init__(self, smooth=False, direct_slicing=True):
        """
        Constructor
        """

        self.smooth = smooth
        self.direct_slicing = direct_slicing


class Profile:
    """
    Contains the beam profile and related quantities including beam spectrum,
    profile derivative.

    Parameters
    ----------

    Beam : object
        Beam from which the profile has to be calculated
    CutOptions : object
        Options for profile cutting (see above)
    FitOptions : object
        Options to get profile position and length (see above)
    FilterOptions : object
        Options to set a filter (see above)
    OtherSlicesOptions : object
        All remaining options, like smooth histogram and direct
        slicing (see above)

    Attributes
    ----------

    Beam : object
    n_slices : int
        number of slices to be used
    cut_left : float
        left extreme of the profile
    cut_right : float
        right extreme of the profile
    n_sigma : float
        defines the left and right extremes of the profile in case those are
        not given explicitly
    edges : float array
        contains the edges of the slices
    bin_centers : float array
        contains the centres of the slices
    bin_size : float
        lenght of one bin (or slice)
    n_macroparticles : float array
        contains the histogram (or profile); its elements are real if the
        smooth histogram tracking is used
    beam_spectrum : float array
        contains the spectrum of the beam (arb. units)
    beam_spectrum_freq : float array
        contains the frequencies on which the spectrum is computed [Hz]
    operations : list
        contains all the methods to be called every turn, like slice track,
        fitting, filtering etc.
    bunchPosition : float
        profile position [s]
    bunchLength : float
        profile length [s]
    filterExtraOptions : unknown (see above)

    Examples
    --------

    >>> n_slices = 100
    >>> CutOptions = profileModule.CutOptions(cut_left=0,
    >>>       cut_right=self.ring.t_rev[0], n_slices = n_slices, cuts_unit='s')
    >>> FitOptions = profileModule.FitOptions(fit_option='gaussian',
    >>>                                        fitExtraOptions=None)
    >>> filter_option = {'pass_frequency':1e7,
    >>>    'stop_frequency':1e8, 'gain_pass':1, 'gain_stop':2,
    >>>    'transfer_function_plot':False}
    >>> FilterOptions = profileModule.FilterOptions(filterMethod='chebishev',
    >>>         filterExtraOptions=filter_option)
    >>> OtherSlicesOptions = profileModule.OtherSlicesOptions(smooth=False,
    >>>                             direct_slicing = True)
    >>> self.profile4 = profileModule.Profile(my_beam, CutOptions = CutOptions,
    >>>                     FitOptions= FitOptions,
    >>>                     FilterOptions=FilterOptions,
    >>>                     OtherSlicesOptions = OtherSlicesOptions)

    """

    def __init__(self, Beam,
                 CutOptions=CutOptions(),
                 FitOptions=FitOptions(),
                 FilterOptions=FilterOptions(),
                 OtherSlicesOptions=OtherSlicesOptions()):
        """
        Constructor
        """

        # Copy of CutOptions object to be usef for reslicing
        self.cut_options = CutOptions

        # Define bins
        CutOptions.set_cuts(Beam)

        # Import (reference) Beam
        self.Beam = Beam

        self.n_slices = 0
        self.cut_left = 0.
        self.cut_right = 0.
        self.n_sigma = 0
        self.edges = None
        self.bin_centers = None
        self.bin_size = 0.
        self.fitExtraOptions = None
        # Get all computed parameters from CutOptions
        self.set_slices_parameters()

        # Initialize profile array as zero array
        self.n_macroparticles = np.zeros(self.n_slices, dtype=bm.precision.real_t, order='C')

        # Initialize beam_spectrum and beam_spectrum_freq as empty arrays
        self.beam_spectrum = np.array([], dtype=bm.precision.real_t, order='C')
        self.beam_spectrum_freq = np.array([], dtype=bm.precision.real_t, order='C')

        if OtherSlicesOptions.smooth:
            self.operations = [self._slice_smooth]
        else:
            self.operations = [self._slice]

        if FitOptions.fit_option is not None:
            self.fit_option = FitOptions.fit_option
            self.bunchPosition = 0.0
            self.bunchLength = 0.0
            if FitOptions.fit_option == 'gaussian':
                self.operations.append(self.apply_fit)
            elif FitOptions.fit_option == 'rms':
                self.operations.append(self.rms)
            elif FitOptions.fit_option == 'fwhm':
                self.operations.append(self.fwhm)

        if FilterOptions.filterMethod == 'chebishev':
            self.filterExtraOptions = FilterOptions.filterExtraOptions
            self.operations.append(self.apply_filter)

        if OtherSlicesOptions.direct_slicing and self.Beam is not None:
            self.track()

        # For CuPy backend
        self._device = 'CPU'

    def set_slices_parameters(self):
        """
        Set various slices parameters.
        """
        self.n_slices, self.cut_left, self.cut_right, self.n_sigma, \
            self.edges, self.bin_centers, self.bin_size = \
            self.cut_options.get_slices_parameters()

    def track(self):
        """
        Track method in order to update the slicing along with the tracker.
        The kwargs are currently only needed to forward the reduce kw argument
        needed for the MPI version.
        """

        for operation in self.operations:
            operation()

    def _slice(self):
        """
        Constant space slicing with a constant frame.
        """
        bm.slice_beam(self.Beam.dt, self.n_macroparticles, self.cut_left,
                 self.cut_right)

        if bm.in_mpi():
            self.reduce_histo()

    def reduce_histo(self, dtype=np.uint32):
        """
        Aggregate all local histograms to calculate the global beam profile.
        """
        if not bm.in_mpi():
            raise RuntimeError(
                'ERROR: Cannot use this routine unless in MPI Mode')

        from ..utils.mpi_config import WORKER

        if WORKER.workers == 1:
            return

        if self.Beam.is_splitted:

            if 'CPU' in bm.device:
                # Convert to uint32t for better performance
                self.n_macroparticles = self.n_macroparticles.astype(
                    dtype, order='C')

            if bm.device == 'GPU':
                import cupy as cp

                # tranfer to cpu
                self.n_macroparticles = cp.asnumpy(self.n_macroparticles, dtype=dtype)

            WORKER.allreduce(self.n_macroparticles)

            if bm.device == 'GPU':
                # transfer back to gpu
                self.n_macroparticles = cp.array(self.n_macroparticles, dtype=bm.precision.real_t)

            if 'CPU' in bm.device:
                # Convert back to float64
                self.n_macroparticles = self.n_macroparticles.astype(
                    dtype=bm.precision.real_t, order='C', copy=False)


    def _slice_smooth(self, reduce=True):
        """
        At the moment 4x slower than _slice but smoother (filtered).
        """
        bm.slice_smooth(self.Beam.dt, self.n_macroparticles, self.cut_left,
                        self.cut_right)

        if bm.in_mpi():
            self.reduce_histo(dtype=np.float64)

    def apply_fit(self):
        """
        It applies Gaussian fit to the profile.
        """

        if self.bunchLength == 0:
            p_0 = [float(self.n_macroparticles.max()),
                  float(self.Beam.dt.mean()),
                  float(self.Beam.dt.std())]
        else:
            p_0 = [float(self.n_macroparticles.max()),
                  float(self.bunchPosition),
                  float(self.bunchLength / 4.)]

        self.fitExtraOptions = ffroutines.gaussian_fit(self.n_macroparticles,
                                                       self.bin_centers, p_0)
        self.bunchPosition = self.fitExtraOptions[1]
        self.bunchLength = 4 * self.fitExtraOptions[2]

    def apply_filter(self):
        """
        It applies Chebishev filter to the profile.
        """
        self.n_macroparticles = ffroutines.beam_profile_filter_chebyshev(
            self.n_macroparticles, self.bin_centers, self.filterExtraOptions)

    def rms(self):
        """
        Computation of the RMS bunch length and position from the line
        density (bunch length = 4sigma).
        """

        self.bunchPosition, self.bunchLength = ffroutines.rms(
            self.n_macroparticles, self.bin_centers)

    def rms_multibunch(self, n_bunches, bunch_spacing_buckets, bucket_size_tau,
                       bucket_tolerance=0.40):
        """
        Computation of the bunch length (4sigma) and position from RMS.
        """

        self.bunchPosition, self.bunchLength = ffroutines.rms_multibunch(
            self.n_macroparticles, self.bin_centers, n_bunches,
            bunch_spacing_buckets, bucket_size_tau, bucket_tolerance)

    def fwhm(self, shift=0):
        """
        Computation of the bunch length and position from the FWHM
        assuming Gaussian line density.
        """

        self.bunchPosition, self.bunchLength = ffroutines.fwhm(
            self.n_macroparticles, self.bin_centers, shift)

    def fwhm_multibunch(self, n_bunches, bunch_spacing_buckets,
                        bucket_size_tau, bucket_tolerance=0.40, shift=0):
        """
        Computation of the bunch length and position from the FWHM
        assuming Gaussian line density for multibunch case.
        """

        self.bunchPosition, self.bunchLength = ffroutines.fwhm_multibunch(
            self.n_macroparticles, self.bin_centers, n_bunches,
            bunch_spacing_buckets, bucket_size_tau, bucket_tolerance, shift)

    def beam_spectrum_freq_generation(self, n_sampling_fft):
        """
        Frequency array of the beam spectrum
        """

        self.beam_spectrum_freq = bm.rfftfreq(n_sampling_fft, self.bin_size)

    def beam_spectrum_generation(self, n_sampling_fft):
        """
        Beam spectrum calculation
        """
        self.beam_spectrum = bm.rfft(self.n_macroparticles, n_sampling_fft)

    def beam_profile_derivative(self, mode='gradient'):
        """
        The input is one of the three available methods for differentiating
        a function. The two outputs are the bin centres and the discrete
        derivative of the Beam profile respectively.*
        """

        bin_centers = self.bin_centers
        dist_centers = bin_centers[1] - bin_centers[0]

        if mode == 'filter1d':
            if bm.device == 'GPU':
                raise RuntimeError('filter1d mode is not supported in GPU.')

            derivative = ndimage.gaussian_filter1d(
                self.n_macroparticles, sigma=1, order=1, mode='wrap') / \
                dist_centers
        elif mode == 'gradient':
            derivative = bm.gradient(self.n_macroparticles, dist_centers)
        elif mode == 'diff':
            derivative = bm.diff(self.n_macroparticles) / dist_centers
            diffCenters = bin_centers[0:-1] + dist_centers / 2
            derivative = bm.interp(bin_centers, diffCenters, derivative)
        else:
            # ProfileDerivativeError
            raise RuntimeError('Option for derivative is not recognized.')

        return bin_centers, derivative

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        # transfer recursively objects to_gpu
        if recursive and self.Beam:
            self.Beam.to_gpu()

        if recursive and self.cut_options:
            self.cut_options.to_gpu()

        import cupy as cp

        self.bin_centers = self.cut_options.bin_centers
        self.edges = self.cut_options.edges

        self.n_macroparticles = cp.array(self.n_macroparticles)
        self.beam_spectrum = cp.array(self.beam_spectrum)
        self.beam_spectrum_freq = cp.array(self.beam_spectrum_freq)

        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        # transfer recursively objects
        if recursive and self.Beam:
            self.Beam.to_cpu()

        if recursive and self.cut_options:
            self.cut_options.to_cpu()

        import cupy as cp

        self.bin_centers = self.cut_options.bin_centers
        self.edges = self.cut_options.edges

        self.n_macroparticles = cp.asnumpy(self.n_macroparticles)
        self.beam_spectrum = cp.asnumpy(self.beam_spectrum)
        self.beam_spectrum_freq = cp.asnumpy(self.beam_spectrum_freq)

        # to make sure it will not be called again
        self._device = 'CPU'
